from decimal import Decimal

import pytest

from pydantic_core import SchemaSerializer, core_schema


def test_decimal():
    v = SchemaSerializer(core_schema.decimal_schema())
    assert v.to_python(Decimal('123.456')) == Decimal('123.456')

    assert v.to_python(Decimal('123.456'), mode='json') == '123.456'
    assert v.to_json(Decimal('123.456')) == b'"123.456"'

    assert v.to_python(Decimal('123456789123456789123456789.123456789123456789123456789')) == Decimal(
        '123456789123456789123456789.123456789123456789123456789'
    )
    assert (
        v.to_json(Decimal('123456789123456789123456789.123456789123456789123456789'))
        == b'"123456789123456789123456789.123456789123456789123456789"'
    )

    with pytest.warns(
        UserWarning, match='Expected `decimal` but got `int` with value `123` - serialized value may not be as expected'
    ):
        assert v.to_python(123, mode='json') == 123

    with pytest.warns(
        UserWarning, match='Expected `decimal` but got `int` with value `123` - serialized value may not be as expected'
    ):
        assert v.to_json(123) == b'123'


def test_decimal_key():
    v = SchemaSerializer(core_schema.dict_schema(core_schema.decimal_schema(), core_schema.decimal_schema()))
    assert v.to_python({Decimal('123.456'): Decimal('123.456')}) == {Decimal('123.456'): Decimal('123.456')}
    assert v.to_python({Decimal('123.456'): Decimal('123.456')}, mode='json') == {'123.456': '123.456'}
    assert v.to_json({Decimal('123.456'): Decimal('123.456')}) == b'{"123.456":"123.456"}'


@pytest.mark.parametrize(
    'value,expected',
    [
        (Decimal('123.456'), '123.456'),
        (Decimal('Infinity'), 'Infinity'),
        (Decimal('-Infinity'), '-Infinity'),
        (Decimal('NaN'), 'NaN'),
    ],
)
def test_decimal_json(value, expected):
    v = SchemaSerializer(core_schema.decimal_schema())
    assert v.to_python(value, mode='json') == expected
    assert v.to_json(value).decode() == f'"{expected}"'


def test_any_decimal_key():
    v = SchemaSerializer(core_schema.dict_schema())
    input_value = {Decimal('123.456'): 1}

    assert v.to_python(input_value, mode='json') == {'123.456': 1}
    assert v.to_json(input_value) == b'{"123.456":1}'



import datetime
import platform
from dataclasses import dataclass
from typing import List, Optional

import pytest
from dirty_equals import AnyThing, HasAttributes, IsList, IsPartialDict, IsStr, IsTuple

import pydantic_core
from pydantic_core import SchemaError, SchemaValidator, ValidationError, core_schema

from ..conftest import Err, plain_repr
from .test_typed_dict import Cls


def test_branch_nullable():
    v = SchemaValidator(
        core_schema.definitions_schema(
            {'type': 'definition-ref', 'schema_ref': 'Branch'},
            [
                {
                    'type': 'typed-dict',
                    'ref': 'Branch',
                    'fields': {
                        'name': {'type': 'typed-dict-field', 'schema': {'type': 'str'}},
                        'sub_branch': {
                            'type': 'typed-dict-field',
                            'schema': {
                                'type': 'default',
                                'schema': {
                                    'type': 'nullable',
                                    'schema': {'type': 'definition-ref', 'schema_ref': 'Branch'},
                                },
                                'default': None,
                            },
                        },
                    },
                }
            ],
        )
    )

    assert v.validate_python({'name': 'root'}) == {'name': 'root', 'sub_branch': None}

    assert v.validate_python({'name': 'root', 'sub_branch': {'name': 'b1'}}) == (
        {'name': 'root', 'sub_branch': {'name': 'b1', 'sub_branch': None}}
    )
    assert v.validate_python({'name': 'root', 'sub_branch': {'name': 'b1', 'sub_branch': {'name': 'b2'}}}) == (
        {'name': 'root', 'sub_branch': {'name': 'b1', 'sub_branch': {'name': 'b2', 'sub_branch': None}}}
    )
    assert ',definitions=[TypedDict(TypedDictValidator{' in plain_repr(v)


def test_unused_ref():
    v = SchemaValidator(
        {
            'type': 'typed-dict',
            'ref': 'Branch',
            'fields': {
                'name': {'type': 'typed-dict-field', 'schema': {'type': 'str'}},
                'other': {'type': 'typed-dict-field', 'schema': {'type': 'int'}},
            },
        }
    )
    assert v.validate_python({'name': 'root', 'other': '4'}) == {'name': 'root', 'other': 4}


def test_nullable_error():
    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.definition_reference_schema('Branch'),
            [
                core_schema.typed_dict_schema(
                    {
                        'width': core_schema.typed_dict_field(core_schema.int_schema()),
                        'sub_branch': core_schema.typed_dict_field(
                            core_schema.with_default_schema(
                                core_schema.union_schema(
                                    [core_schema.none_schema(), core_schema.definition_reference_schema('Branch')]
                                ),
                                default=None,
                            )
                        ),
                    },
                    ref='Branch',
                )
            ],
        )
    )
    assert v.validate_python({'width': 123, 'sub_branch': {'width': 321}}) == (
        {'width': 123, 'sub_branch': {'width': 321, 'sub_branch': None}}
    )
    with pytest.raises(ValidationError) as exc_info:
        assert v.validate_python({'width': 123, 'sub_branch': {'width': 'wrong'}})
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'none_required',
            'loc': ('sub_branch', 'none'),
            'msg': 'Input should be None',
            'input': {'width': 'wrong'},
        },
        {
            'type': 'int_parsing',
            'loc': ('sub_branch', 'typed-dict', 'width'),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'wrong',
        },
    ]


def test_list():
    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.definition_reference_schema('BranchList'),
            [
                core_schema.typed_dict_schema(
                    {
                        'width': core_schema.typed_dict_field(core_schema.int_schema()),
                        'branches': core_schema.typed_dict_field(
                            core_schema.with_default_schema(
                                core_schema.list_schema(core_schema.definition_reference_schema('BranchList')),
                                default=None,
                            )
                        ),
                    },
                    ref='BranchList',
                )
            ],
        )
    )
    assert v.validate_python({'width': 1, 'branches': [{'width': 2}, {'width': 3, 'branches': [{'width': 4}]}]}) == (
        {
            'width': 1,
            'branches': [{'width': 2, 'branches': None}, {'width': 3, 'branches': [{'width': 4, 'branches': None}]}],
        }
    )
    assert ',definitions=[TypedDict(TypedDictValidator{' in plain_repr(v)


def test_multiple_intertwined():
    """
    like:
    from typing import List, Optional
    class Foo:
        height: int
        class Bar:
            width: int
            bars: List['Bar']
            foo: Optional['Foo']
        bar = Bar
    """

    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.definition_reference_schema('Foo'),
            [
                core_schema.typed_dict_schema(
                    {
                        'height': core_schema.typed_dict_field(core_schema.int_schema()),
                        'bar': core_schema.typed_dict_field(core_schema.definition_reference_schema('Bar')),
                    },
                    ref='Foo',
                ),
                core_schema.typed_dict_schema(
                    {
                        'width': core_schema.typed_dict_field(core_schema.int_schema()),
                        'bars': core_schema.typed_dict_field(
                            core_schema.with_default_schema(
                                core_schema.list_schema(core_schema.definition_reference_schema('Bar')), default=None
                            )
                        ),
                        'foo': core_schema.typed_dict_field(
                            core_schema.with_default_schema(
                                core_schema.union_schema(
                                    [core_schema.none_schema(), core_schema.definition_reference_schema('Foo')]
                                ),
                                default=None,
                            )
                        ),
                    },
                    ref='Bar',
                ),
            ],
        )
    )
    v.validate_python(
        {
            'height': 1,
            'bar': {
                'width': 2,
                'bars': [{'width': 3}],
                'foo': {'height': 4, 'bar': {'width': 5, 'bars': [], 'foo': None}},
            },
        }
    )


def test_model_class():
    class Branch:
        # this is not required, but it avoids `__pydantic_fields_set__` being included in `__dict__`
        __slots__ = '__dict__', '__pydantic_fields_set__', '__pydantic_extra__', '__pydantic_private__'
        # these are here just as decoration
        width: int
        branch: Optional['Branch']

    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.definition_reference_schema('Branch'),
            [
                core_schema.model_schema(
                    Branch,
                    core_schema.model_fields_schema(
                        {
                            'width': core_schema.model_field(core_schema.int_schema()),
                            'branch': core_schema.model_field(
                                core_schema.with_default_schema(
                                    core_schema.union_schema(
                                        [core_schema.none_schema(), core_schema.definition_reference_schema('Branch')]
                                    ),
                                    default=None,
                                )
                            ),
                        }
                    ),
                    ref='Branch',
                )
            ],
        )
    )
    m1: Branch = v.validate_python({'width': '1'})
    assert isinstance(m1, Branch)
    assert m1.__pydantic_fields_set__ == {'width'}
    assert m1.__dict__ == {'width': 1, 'branch': None}
    assert m1.width == 1
    assert m1.branch is None

    m2: Branch = v.validate_python({'width': '10', 'branch': {'width': 20}})
    assert isinstance(m2, Branch)
    assert m2.__pydantic_fields_set__ == {'width', 'branch'}
    assert m2.width == 10
    assert isinstance(m2.branch, Branch)
    assert m2.branch.width == 20
    assert m2.branch.branch is None


def test_invalid_schema():
    with pytest.raises(SchemaError, match='Definitions error: definition `Branch` was never filled'):
        SchemaValidator(
            {
                'type': 'list',
                'items_schema': {
                    'type': 'typed-dict',
                    'fields': {
                        'width': {'type': 'typed-dict-field', 'schema': {'type': 'int'}},
                        'branch': {
                            'type': 'typed-dict-field',
                            'schema': {
                                'type': 'default',
                                'schema': {
                                    'type': 'nullable',
                                    'schema': {'type': 'definition-ref', 'schema_ref': 'Branch'},
                                },
                                'default': None,
                            },
                        },
                    },
                },
            }
        )


def test_outside_parent():
    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.typed_dict_schema(
                {
                    'tuple1': core_schema.typed_dict_field(core_schema.definition_reference_schema('tuple-iis')),
                    'tuple2': core_schema.typed_dict_field(core_schema.definition_reference_schema('tuple-iis')),
                }
            ),
            [
                core_schema.tuple_schema(
                    [core_schema.int_schema(), core_schema.int_schema(), core_schema.str_schema()], ref='tuple-iis'
                )
            ],
        )
    )

    assert v.validate_python({'tuple1': [1, '1', 'frog'], 'tuple2': [2, '2', 'toad']}) == {
        'tuple1': (1, 1, 'frog'),
        'tuple2': (2, 2, 'toad'),
    }


def test_recursion_branch():
    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.definition_reference_schema('Branch'),
            [
                core_schema.typed_dict_schema(
                    {
                        'name': core_schema.typed_dict_field(core_schema.str_schema()),
                        'branch': core_schema.typed_dict_field(
                            core_schema.with_default_schema(
                                core_schema.nullable_schema(core_schema.definition_reference_schema('Branch')),
                                default=None,
                            )
                        ),
                    },
                    ref='Branch',
                )
            ],
        ),
        {'from_attributes': True},
    )
    assert ',definitions=[TypedDict(TypedDictValidator{' in plain_repr(v)

    assert v.validate_python({'name': 'root'}) == {'name': 'root', 'branch': None}
    assert v.validate_python({'name': 'root', 'branch': {'name': 'b1', 'branch': None}}) == {
        'name': 'root',
        'branch': {'name': 'b1', 'branch': None},
    }

    b = {'name': 'recursive'}
    b['branch'] = b
    with pytest.raises(ValidationError) as exc_info:
        assert v.validate_python(b)
    assert exc_info.value.title == 'typed-dict'
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'recursion_loop',
            'loc': ('branch',),
            'msg': 'Recursion error - cyclic reference detected',
            'input': {'name': 'recursive', 'branch': IsPartialDict(name='recursive')},
        }
    ]


def test_recursion_branch_from_attributes():
    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.definition_reference_schema('Branch'),
            [
                core_schema.model_fields_schema(
                    {
                        'name': core_schema.model_field(core_schema.str_schema()),
                        'branch': core_schema.model_field(
                            core_schema.with_default_schema(
                                core_schema.nullable_schema(core_schema.definition_reference_schema('Branch')),
                                default=None,
                            )
                        ),
                    },
                    ref='Branch',
                )
            ],
        ),
        {'from_attributes': True},
    )

    assert v.validate_python({'name': 'root'}) == ({'name': 'root', 'branch': None}, None, {'name'})
    model_dict, model_extra, fields_set = v.validate_python({'name': 'root', 'branch': {'name': 'b1', 'branch': None}})
    assert model_dict == {'name': 'root', 'branch': ({'name': 'b1', 'branch': None}, None, {'name', 'branch'})}
    assert model_extra is None
    assert fields_set == {'name', 'branch'}

    data = Cls(name='root')
    data.branch = Cls(name='b1', branch=None)
    model_dict, model_extra, fields_set = v.validate_python(data)
    assert model_dict == {'name': 'root', 'branch': ({'name': 'b1', 'branch': None}, None, {'name', 'branch'})}
    assert model_extra is None
    assert fields_set == {'name', 'branch'}

    data = Cls(name='root')
    data.branch = data
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(data)
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'recursion_loop',
            'loc': ('branch',),
            'msg': 'Recursion error - cyclic reference detected',
            'input': HasAttributes(name='root', branch=AnyThing()),
        }
    ]


def test_definition_list():
    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.definition_reference_schema('the-list'),
            [core_schema.list_schema(core_schema.definition_reference_schema('the-list'), ref='the-list')],
        )
    )
    assert ',definitions=[List(ListValidator{' in plain_repr(v)
    assert v.validate_python([]) == []
    assert v.validate_python([[]]) == [[]]

    data = list()
    data.append(data)
    with pytest.raises(ValidationError) as exc_info:
        assert v.validate_python(data)
    assert exc_info.value.title == 'list[...]'
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'recursion_loop',
            'loc': (0,),
            'msg': 'Recursion error - cyclic reference detected',
            'input': [IsList(length=1)],
        }
    ]


@pytest.fixture(scope='module')
def multiple_tuple_schema() -> SchemaValidator:
    return SchemaValidator(
        core_schema.definitions_schema(
            core_schema.typed_dict_schema(
                {
                    'f1': core_schema.typed_dict_field(core_schema.definition_reference_schema('t')),
                    'f2': core_schema.typed_dict_field(
                        core_schema.with_default_schema(
                            core_schema.nullable_schema(core_schema.definition_reference_schema('t')), default=None
                        )
                    ),
                }
            ),
            [
                core_schema.tuple_schema(
                    [
                        core_schema.int_schema(),
                        core_schema.nullable_schema(core_schema.definition_reference_schema('t')),
                    ],
                    ref='t',
                )
            ],
        )
    )


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ({'f1': [1, None]}, {'f1': (1, None), 'f2': None}),
        ({'f1': [1, None], 'f2': [2, None]}, {'f1': (1, None), 'f2': (2, None)}),
        (
            {'f1': [1, (3, None)], 'f2': [2, (4, (4, (5, None)))]},
            {'f1': (1, (3, None)), 'f2': (2, (4, (4, (5, None))))},
        ),
        ({'f1': [1, 2]}, Err(r'f1.1\s+Input should be a valid tuple')),
        ({'f1': [1, (3, None)], 'f2': [2, (4, (4, (5, 6)))]}, Err(r'f2.1.1.1.1\s+Input should be a valid tuple')),
    ],
)
def test_multiple_tuple_param(multiple_tuple_schema: SchemaValidator, input_value, expected):
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=expected.message):
            multiple_tuple_schema.validate_python(input_value)
        # debug(repr(exc_info.value))
    else:
        assert multiple_tuple_schema.validate_python(input_value) == expected


def test_multiple_tuple_repeat(multiple_tuple_schema: SchemaValidator):
    t = (42, None)
    assert multiple_tuple_schema.validate_python({'f1': (1, t), 'f2': (2, t)}) == {
        'f1': (1, (42, None)),
        'f2': (2, (42, None)),
    }


def test_multiple_tuple_recursion(multiple_tuple_schema: SchemaValidator):
    data = [1]
    data.append(data)
    with pytest.raises(ValidationError) as exc_info:
        multiple_tuple_schema.validate_python({'f1': data, 'f2': data})

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'recursion_loop',
            'loc': ('f1', 1),
            'msg': 'Recursion error - cyclic reference detected',
            'input': [1, IsList(length=2)],
        },
        {
            'type': 'recursion_loop',
            'loc': ('f2', 1),
            'msg': 'Recursion error - cyclic reference detected',
            'input': [1, IsList(length=2)],
        },
    ]


def test_multiple_tuple_recursion_once(multiple_tuple_schema: SchemaValidator):
    data = [1]
    data.append(data)
    with pytest.raises(ValidationError) as exc_info:
        multiple_tuple_schema.validate_python({'f1': data, 'f2': data})

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'recursion_loop',
            'loc': ('f1', 1),
            'msg': 'Recursion error - cyclic reference detected',
            'input': [1, IsList(length=2)],
        },
        {
            'type': 'recursion_loop',
            'loc': ('f2', 1),
            'msg': 'Recursion error - cyclic reference detected',
            'input': [1, IsList(length=2)],
        },
    ]


def test_definition_wrap():
    def wrap_func(input_value, validator, info):
        return validator(input_value) + (42,)

    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.definition_reference_schema('wrapper'),
            [
                core_schema.with_info_wrap_validator_function(
                    wrap_func,
                    core_schema.tuple_schema(
                        [
                            core_schema.int_schema(),
                            core_schema.nullable_schema(core_schema.definition_reference_schema('wrapper')),
                        ]
                    ),
                    ref='wrapper',
                )
            ],
        )
    )
    assert v.validate_python((1, None)) == (1, None, 42)
    assert v.validate_python((1, (2, (3, None)))) == (1, (2, (3, None, 42), 42), 42)
    t = [1]
    t.append(t)
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(t)
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'recursion_loop',
            'loc': (1,),
            'msg': 'Recursion error - cyclic reference detected',
            'input': IsList(positions={0: 1}, length=2),
        }
    ]


def test_union_ref_strictness():
    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.typed_dict_schema(
                {
                    'a': core_schema.typed_dict_field(core_schema.definition_reference_schema('int-type')),
                    'b': core_schema.typed_dict_field(
                        core_schema.union_schema(
                            [core_schema.definition_reference_schema('int-type'), core_schema.str_schema()]
                        )
                    ),
                }
            ),
            [core_schema.int_schema(ref='int-type')],
        )
    )
    assert v.validate_python({'a': 1, 'b': '2'}) == {'a': 1, 'b': '2'}
    assert v.validate_python({'a': 1, 'b': 2}) == {'a': 1, 'b': 2}

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'a': 1, 'b': []})

    assert exc_info.value.errors(include_url=False) == [
        {'type': 'int_type', 'loc': ('b', 'int'), 'msg': 'Input should be a valid integer', 'input': []},
        {'type': 'string_type', 'loc': ('b', 'str'), 'msg': 'Input should be a valid string', 'input': []},
    ]


def test_union_container_strictness():
    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.typed_dict_schema(
                {
                    'b': core_schema.typed_dict_field(
                        core_schema.union_schema(
                            [core_schema.definition_reference_schema('int-type'), core_schema.str_schema()]
                        )
                    ),
                    'a': core_schema.typed_dict_field(core_schema.definition_reference_schema('int-type')),
                }
            ),
            [core_schema.int_schema(ref='int-type')],
        )
    )
    assert v.validate_python({'a': 1, 'b': '2'}) == {'a': 1, 'b': '2'}
    assert v.validate_python({'a': 1, 'b': 2}) == {'a': 1, 'b': 2}

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'a': 1, 'b': []})

    assert exc_info.value.errors(include_url=False) == [
        {'type': 'int_type', 'loc': ('b', 'int'), 'msg': 'Input should be a valid integer', 'input': []},
        {'type': 'string_type', 'loc': ('b', 'str'), 'msg': 'Input should be a valid string', 'input': []},
    ]


@pytest.mark.parametrize('strict', [True, False], ids=lambda s: f'strict={s}')
def test_union_cycle(strict: bool):
    s = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.definition_reference_schema('root-schema'),
            [
                core_schema.union_schema(
                    [
                        core_schema.typed_dict_schema(
                            {
                                'foobar': core_schema.typed_dict_field(
                                    core_schema.list_schema(core_schema.definition_reference_schema('root-schema'))
                                )
                            }
                        )
                    ],
                    auto_collapse=False,
                    strict=strict,
                    ref='root-schema',
                )
            ],
        )
    )

    data = {'foobar': []}
    data['foobar'].append(data)

    with pytest.raises(ValidationError) as exc_info:
        s.validate_python(data)
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'recursion_loop',
            'loc': ('typed-dict', 'foobar', 0),
            'msg': 'Recursion error - cyclic reference detected',
            'input': {'foobar': [{'foobar': IsList(length=1)}]},
        }
    ]


def test_function_name():
    def f(input_value, info):
        return input_value + ' Changed'

    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.definition_reference_schema('root-schema'),
            [
                core_schema.union_schema(
                    [
                        core_schema.with_info_after_validator_function(
                            f, core_schema.definition_reference_schema('root-schema')
                        ),
                        core_schema.int_schema(),
                    ],
                    ref='root-schema',
                )
            ],
        )
    )

    assert v.validate_python(123) == 123

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python('input value')

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'recursion_loop',
            'loc': ('function-after[f(), ...]',),
            'msg': 'Recursion error - cyclic reference detected',
            'input': 'input value',
        },
        {
            'type': 'int_parsing',
            'loc': ('int',),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'input value',
        },
    ]


@pytest.mark.skipif(
    platform.python_implementation() == 'PyPy' and pydantic_core._pydantic_core.build_profile == 'debug',
    reason='PyPy does not have enough stack space for Rust debug builds to recurse very deep',
)
@pytest.mark.parametrize('strict', [True, False], ids=lambda s: f'strict={s}')
def test_function_change_id(strict: bool):
    def f(input_value, info):
        _, count = input_value.split('-')
        return f'f-{int(count) + 1}'

    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.definition_reference_schema('root-schema'),
            [
                core_schema.union_schema(
                    [
                        core_schema.with_info_before_validator_function(
                            f, core_schema.definition_reference_schema('root-schema')
                        )
                    ],
                    auto_collapse=False,
                    strict=strict,
                    ref='root-schema',
                )
            ],
        )
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python('start-0')

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'recursion_loop',
            'loc': IsTuple(length=(1, 255)),
            'msg': 'Recursion error - cyclic reference detected',
            'input': IsStr(regex=r'f-\d+'),
        }
    ]


def test_many_uses_of_ref():
    # check we can safely exceed RECURSION_GUARD_LIMIT without upsetting the recursion guard
    v = SchemaValidator(
        core_schema.definitions_schema(
            core_schema.definition_reference_schema('Branch'),
            [
                core_schema.typed_dict_schema(
                    {
                        'name': core_schema.typed_dict_field(core_schema.definition_reference_schema('limited-string')),
                        'other_names': core_schema.typed_dict_field(
                            core_schema.list_schema(core_schema.definition_reference_schema('limited-string'))
                        ),
                    },
                    ref='Branch',
                ),
                core_schema.str_schema(max_length=8, ref='limited-string'),
            ],
        )
    )

    assert v.validate_python({'name': 'Anne', 'other_names': ['Bob', 'Charlie']}) == {
        'name': 'Anne',
        'other_names': ['Bob', 'Charlie'],
    }

    with pytest.raises(ValidationError, match=r'other_names.2\s+String should have at most 8 characters'):
        v.validate_python({'name': 'Anne', 'other_names': ['Bob', 'Charlie', 'Daveeeeee']})

    long_input = {'name': 'Anne', 'other_names': [f'p-{i}' for i in range(300)]}
    assert v.validate_python(long_input) == long_input


def test_error_inside_definition_wrapper():
    with pytest.raises(SchemaError) as exc_info:
        SchemaValidator(
            {
                'type': 'typed-dict',
                'ref': 'Branch',
                'fields': {
                    'sub_branch': {
                        'type': 'typed-dict-field',
                        'schema': {
                            'type': 'default',
                            'schema': {
                                'type': 'nullable',
                                'schema': {'type': 'definition-ref', 'schema_ref': 'Branch'},
                            },
                            'default': None,
                            'default_factory': lambda x: 'foobar',
                        },
                    }
                },
            }
        )
    assert str(exc_info.value) == (
        'Error building "typed-dict" validator:\n'
        '  SchemaError: Field "sub_branch":\n'
        '  SchemaError: Error building "default" validator:\n'
        "  SchemaError: 'default' and 'default_factory' cannot be used together"
    )


def test_recursive_definitions_schema(pydantic_version) -> None:
    s = core_schema.definitions_schema(
        core_schema.definition_reference_schema('a'),
        [
            core_schema.typed_dict_schema(
                {
                    'b': core_schema.typed_dict_field(
                        core_schema.list_schema(core_schema.definition_reference_schema('b'))
                    )
                },
                ref='a',
            ),
            core_schema.typed_dict_schema(
                {
                    'a': core_schema.typed_dict_field(
                        core_schema.list_schema(core_schema.definition_reference_schema('a'))
                    )
                },
                ref='b',
            ),
        ],
    )

    v = SchemaValidator(s)

    assert v.validate_python({'b': [{'a': []}]}) == {'b': [{'a': []}]}

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'b': [{'a': {}}]})

    assert exc_info.value.errors() == [
        {
            'type': 'list_type',
            'loc': ('b', 0, 'a'),
            'msg': 'Input should be a valid list',
            'input': {},
            'url': f'https://errors.pydantic.dev/{pydantic_version}/v/list_type',
        }
    ]


def test_unsorted_definitions_schema() -> None:
    s = core_schema.definitions_schema(
        core_schema.definition_reference_schema('td'),
        [
            core_schema.typed_dict_schema(
                {'x': core_schema.typed_dict_field(core_schema.definition_reference_schema('int'))}, ref='td'
            ),
            core_schema.int_schema(ref='int'),
        ],
    )

    v = SchemaValidator(s)

    assert v.validate_python({'x': 123}) == {'x': 123}

    with pytest.raises(ValidationError):
        v.validate_python({'x': 'abc'})


def test_validate_assignment(pydantic_version) -> None:
    @dataclass
    class Model:
        x: List['Model']

    schema = core_schema.definitions_schema(
        core_schema.definition_reference_schema('model'),
        [
            core_schema.dataclass_schema(
                Model,
                core_schema.dataclass_args_schema(
                    'Model',
                    [
                        core_schema.dataclass_field(
                            name='x',
                            schema=core_schema.list_schema(core_schema.definition_reference_schema('model')),
                            kw_only=False,
                        )
                    ],
                ),
                ['x'],
                ref='model',
                config=core_schema.CoreConfig(revalidate_instances='always'),
            )
        ],
    )

    v = SchemaValidator(schema)

    data = [Model(x=[Model(x=[])])]
    instance = Model(x=[])
    v.validate_assignment(instance, 'x', data)
    assert instance.x == data

    with pytest.raises(ValidationError) as exc_info:
        v.validate_assignment(instance, 'x', [Model(x=[Model(x=[Model(x=[123])])])])

    assert exc_info.value.errors() == [
        {
            'type': 'dataclass_type',
            'loc': ('x', 0, 'x', 0, 'x', 0, 'x', 0),
            'msg': 'Input should be a dictionary or an instance of Model',
            'input': 123,
            'ctx': {'class_name': 'Model'},
            'url': f'https://errors.pydantic.dev/{pydantic_version}/v/dataclass_type',
        }
    ]


def test_cyclic_data() -> None:
    cyclic_data = {}
    cyclic_data['b'] = {'a': cyclic_data}

    schema = core_schema.definitions_schema(
        core_schema.definition_reference_schema('a'),
        [
            core_schema.typed_dict_schema(
                {
                    'b': core_schema.typed_dict_field(
                        core_schema.nullable_schema(core_schema.definition_reference_schema('b'))
                    )
                },
                ref='a',
            ),
            core_schema.typed_dict_schema(
                {
                    'a': core_schema.typed_dict_field(
                        core_schema.nullable_schema(core_schema.definition_reference_schema('a'))
                    )
                },
                ref='b',
            ),
        ],
    )

    validator = SchemaValidator(schema)

    with pytest.raises(ValidationError) as exc_info:
        validator.validate_python(cyclic_data)

    assert exc_info.value.title == 'typed-dict'
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'recursion_loop',
            'loc': ('b', 'a'),
            'msg': 'Recursion error - cyclic reference detected',
            'input': cyclic_data,
        }
    ]


def test_cyclic_data_threeway() -> None:
    cyclic_data = {}
    cyclic_data['b'] = {'c': {'a': cyclic_data}}

    schema = core_schema.definitions_schema(
        core_schema.definition_reference_schema('a'),
        [
            core_schema.typed_dict_schema(
                {
                    'b': core_schema.typed_dict_field(
                        core_schema.nullable_schema(core_schema.definition_reference_schema('b'))
                    )
                },
                ref='a',
            ),
            core_schema.typed_dict_schema(
                {
                    'c': core_schema.typed_dict_field(
                        core_schema.nullable_schema(core_schema.definition_reference_schema('c'))
                    )
                },
                ref='b',
            ),
            core_schema.typed_dict_schema(
                {
                    'a': core_schema.typed_dict_field(
                        core_schema.nullable_schema(core_schema.definition_reference_schema('a'))
                    )
                },
                ref='c',
            ),
        ],
    )

    validator = SchemaValidator(schema)

    with pytest.raises(ValidationError) as exc_info:
        validator.validate_python(cyclic_data)

    assert exc_info.value.title == 'typed-dict'
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'recursion_loop',
            'loc': ('b', 'c', 'a'),
            'msg': 'Recursion error - cyclic reference detected',
            'input': cyclic_data,
        }
    ]


def test_complex_recursive_type() -> None:
    schema = core_schema.definitions_schema(
        core_schema.definition_reference_schema('JsonType'),
        [
            core_schema.nullable_schema(
                core_schema.union_schema(
                    [
                        core_schema.list_schema(core_schema.definition_reference_schema('JsonType')),
                        core_schema.dict_schema(
                            core_schema.str_schema(), core_schema.definition_reference_schema('JsonType')
                        ),
                        core_schema.str_schema(),
                        core_schema.int_schema(),
                        core_schema.float_schema(),
                        core_schema.bool_schema(),
                    ]
                ),
                ref='JsonType',
            )
        ],
    )

    validator = SchemaValidator(schema)

    with pytest.raises(ValidationError) as exc_info:
        validator.validate_python({'a': datetime.date(year=1992, month=12, day=11)})

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'list_type',
            'loc': ('list[nullable[union[list[...],dict[str,...],str,int,float,bool]]]',),
            'msg': 'Input should be a valid list',
            'input': {'a': datetime.date(1992, 12, 11)},
        },
        {
            'type': 'list_type',
            'loc': ('dict[str,...]', 'a', 'list[nullable[union[list[...],dict[str,...],str,int,float,bool]]]'),
            'msg': 'Input should be a valid list',
            'input': datetime.date(1992, 12, 11),
        },
        {
            'type': 'dict_type',
            'loc': ('dict[str,...]', 'a', 'dict[str,...]'),
            'msg': 'Input should be a valid dictionary',
            'input': datetime.date(1992, 12, 11),
        },
        {
            'type': 'string_type',
            'loc': ('dict[str,...]', 'a', 'str'),
            'msg': 'Input should be a valid string',
            'input': datetime.date(1992, 12, 11),
        },
        {
            'type': 'int_type',
            'loc': ('dict[str,...]', 'a', 'int'),
            'msg': 'Input should be a valid integer',
            'input': datetime.date(1992, 12, 11),
        },
        {
            'type': 'float_type',
            'loc': ('dict[str,...]', 'a', 'float'),
            'msg': 'Input should be a valid number',
            'input': datetime.date(1992, 12, 11),
        },
        {
            'type': 'bool_type',
            'loc': ('dict[str,...]', 'a', 'bool'),
            'msg': 'Input should be a valid boolean',
            'input': datetime.date(1992, 12, 11),
        },
        {
            'type': 'string_type',
            'loc': ('str',),
            'msg': 'Input should be a valid string',
            'input': {'a': datetime.date(1992, 12, 11)},
        },
        {
            'type': 'int_type',
            'loc': ('int',),
            'msg': 'Input should be a valid integer',
            'input': {'a': datetime.date(1992, 12, 11)},
        },
        {
            'type': 'float_type',
            'loc': ('float',),
            'msg': 'Input should be a valid number',
            'input': {'a': datetime.date(1992, 12, 11)},
        },
        {
            'type': 'bool_type',
            'loc': ('bool',),
            'msg': 'Input should be a valid boolean',
            'input': {'a': datetime.date(1992, 12, 11)},
        },
    ]


def test_no_exponential_blowup():
    """See https://github.com/pydantic/pydantic/issues/8049

    There was a performance bug which led to exponential blowup when trying to
    build a schema with many intermingled recursive definitions.
    """
    unions = core_schema.union_schema([core_schema.definition_reference_schema(f'foo_{i}') for i in range(100)])

    schema = core_schema.definitions_schema(
        core_schema.typed_dict_schema({'x': core_schema.typed_dict_field(unions)}),
        definitions=[
            core_schema.typed_dict_schema({'a': core_schema.typed_dict_field(unions)}, ref=f'foo_{i}')
            for i in range(100)
        ],
    )

    SchemaValidator(schema)



from dataclasses import dataclass
from enum import Enum
from typing import Literal, Union

import pytest

from pydantic_core import SchemaError, SchemaSerializer, core_schema

from ..conftest import plain_repr


def test_int_literal():
    s = SchemaSerializer(core_schema.literal_schema([1, 2, 3]))
    r = plain_repr(s)
    assert 'expected_int:{' in r
    assert 'expected_str:{}' in r
    assert 'expected_py:None' in r

    assert s.to_python(1) == 1
    assert s.to_python(1, mode='json') == 1
    assert s.to_python(44) == 44
    assert s.to_json(1) == b'1'

    # with pytest.warns(UserWarning, match='Expected `int` but got `str` - serialized value may not be as expected'):
    assert s.to_python('a', mode='json') == 'a'

    # with pytest.warns(UserWarning, match='Expected `int` but got `str` - serialized value may not be as expected'):
    assert s.to_json('a') == b'"a"'


def test_str_literal():
    s = SchemaSerializer(core_schema.literal_schema(['a', 'b', 'c']))
    r = plain_repr(s)
    assert 'expected_str:{' in r
    assert 'expected_int:{}' in r
    assert 'expected_py:None' in r

    assert s.to_python('a') == 'a'
    assert s.to_python('a', mode='json') == 'a'
    assert s.to_python('not in literal') == 'not in literal'
    assert s.to_json('a') == b'"a"'

    # with pytest.warns(UserWarning, match='Expected `str` but got `int` - serialized value may not be as expected'):
    assert s.to_python(1, mode='json') == 1

    # with pytest.warns(UserWarning, match='Expected `str` but got `int` - serialized value may not be as expected'):
    assert s.to_json(1) == b'1'


def test_other_literal():
    s = SchemaSerializer(core_schema.literal_schema(['a', 1]))
    assert 'expected_int:{1},expected_str:{"a"},expected_py:None' in plain_repr(s)

    assert s.to_python('a') == 'a'
    assert s.to_python('a', mode='json') == 'a'
    assert s.to_python('not in literal') == 'not in literal'
    assert s.to_json('a') == b'"a"'

    assert s.to_python(1) == 1
    assert s.to_python(1, mode='json') == 1
    assert s.to_python(44) == 44
    assert s.to_json(1) == b'1'


def test_empty_literal():
    with pytest.raises(SchemaError, match='`expected` should have length > 0'):
        SchemaSerializer(core_schema.literal_schema([]))


def test_bool_literal():
    s = SchemaSerializer(core_schema.literal_schema([False]))
    assert 'expected_int:{},expected_str:{},expected_py:Some(Py(' in plain_repr(s)

    assert s.to_python(False) is False
    assert s.to_python(False, mode='json') is False
    assert s.to_python(True) is True
    assert s.to_json(False) == b'false'


def test_literal_with_enum() -> None:
    class SomeEnum(str, Enum):
        CAT = 'cat'
        DOG = 'dog'

    @dataclass
    class Dog:
        name: str
        type: Literal[SomeEnum.DOG] = SomeEnum.DOG

    @dataclass
    class Cat:
        name: str
        type: Literal[SomeEnum.CAT] = SomeEnum.CAT

    @dataclass
    class Yard:
        pet: Union[Dog, Cat]

    serializer = SchemaSerializer(
        core_schema.model_schema(
            cls=Yard,
            schema=core_schema.model_fields_schema(
                fields={
                    'pet': core_schema.model_field(
                        schema=core_schema.tagged_union_schema(
                            choices={
                                SomeEnum.DOG: core_schema.model_schema(
                                    cls=Dog,
                                    schema=core_schema.model_fields_schema(
                                        fields={
                                            'type': core_schema.model_field(
                                                schema=core_schema.with_default_schema(
                                                    schema=core_schema.literal_schema([SomeEnum.DOG]),
                                                    default=SomeEnum.DOG,
                                                )
                                            ),
                                            'name': core_schema.model_field(schema=core_schema.str_schema()),
                                        },
                                        model_name='Dog',
                                    ),
                                ),
                                SomeEnum.CAT: core_schema.model_schema(
                                    cls=Cat,
                                    schema=core_schema.model_fields_schema(
                                        fields={
                                            'type': core_schema.model_field(
                                                schema=core_schema.with_default_schema(
                                                    schema=core_schema.literal_schema([SomeEnum.CAT]),
                                                    default=SomeEnum.CAT,
                                                )
                                            ),
                                            'name': core_schema.model_field(schema=core_schema.str_schema()),
                                        },
                                        model_name='Cat',
                                    ),
                                ),
                            },
                            discriminator='type',
                            strict=False,
                            from_attributes=True,
                        )
                    )
                }
            ),
        )
    )

    yard = Yard(pet=Dog(name='Rex'))
    assert serializer.to_python(yard, mode='json') == {'pet': {'type': 'dog', 'name': 'Rex'}}



import pytest

from pydantic_core import SchemaSerializer, core_schema


def test_nullable():
    s = SchemaSerializer(core_schema.nullable_schema(core_schema.int_schema()))
    assert s.to_python(None) is None
    assert s.to_python(1) == 1
    assert s.to_python(None, mode='json') is None
    assert s.to_python(1, mode='json') == 1
    assert s.to_json(1) == b'1'
    assert s.to_json(None) == b'null'
    with pytest.warns(
        UserWarning, match="Expected `int` but got `str` with value `'aaa'` - serialized value may not be as expected"
    ):
        assert s.to_json('aaa') == b'"aaa"'



import math
import re
import sys
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Mapping, Union

import pytest
from dirty_equals import FunctionCheck, HasRepr, IsStr

from pydantic_core import CoreConfig, SchemaError, SchemaValidator, ValidationError, core_schema, validate_core_schema

from ..conftest import Err, PyAndJson


class Cls:
    def __init__(self, **attributes):
        for k, v in attributes.items():
            setattr(self, k, v)

    def __repr__(self):
        return 'Cls({})'.format(', '.join(f'{k}={v!r}' for k, v in self.__dict__.items()))


class Map(Mapping):
    def __init__(self, **kwargs):
        self._d = kwargs

    def __iter__(self):
        return iter(self._d)

    def __len__(self) -> int:
        return len(self._d)

    def __getitem__(self, k, /):
        return self._d[k]

    def __repr__(self):
        return 'Map({})'.format(', '.join(f'{k}={v!r}' for k, v in self._d.items()))


def test_simple():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {'type': 'model-field', 'schema': {'type': 'str'}},
                'field_b': {'type': 'model-field', 'schema': {'type': 'int'}},
            },
        }
    )

    assert v.validate_python({'field_a': b'abc', 'field_b': 1}) == (
        {'field_a': 'abc', 'field_b': 1},
        None,
        {'field_a', 'field_b'},
    )


def test_strict():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {'type': 'model-field', 'schema': {'type': 'str'}},
                'field_b': {'type': 'model-field', 'schema': {'type': 'int'}},
            },
        },
        {'strict': True},
    )

    assert v.validate_python({'field_a': 'hello', 'field_b': 12}) == (
        {'field_a': 'hello', 'field_b': 12},
        None,
        {'field_a', 'field_b'},
    )

    with pytest.raises(ValidationError) as exc_info:
        assert v.validate_python({'field_a': 123, 'field_b': '123'})
    assert exc_info.value.errors(include_url=False) == [
        {'type': 'string_type', 'loc': ('field_a',), 'msg': 'Input should be a valid string', 'input': 123},
        {'type': 'int_type', 'loc': ('field_b',), 'msg': 'Input should be a valid integer', 'input': '123'},
    ]


def test_with_default():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {'type': 'model-field', 'schema': {'type': 'str'}},
                'field_b': {
                    'type': 'model-field',
                    'schema': {'type': 'default', 'schema': {'type': 'int'}, 'default': 666},
                },
            },
        }
    )

    assert v.validate_python({'field_a': b'abc'}) == ({'field_a': 'abc', 'field_b': 666}, None, {'field_a'})
    assert v.validate_python({'field_a': b'abc', 'field_b': 1}) == (
        {'field_a': 'abc', 'field_b': 1},
        None,
        {'field_b', 'field_a'},
    )


def test_missing_error(pydantic_version):
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {'type': 'model-field', 'schema': {'type': 'str'}},
                'field_b': {'type': 'model-field', 'schema': {'type': 'int'}},
            },
        }
    )
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'field_a': b'abc'})
    assert (
        str(exc_info.value)
        == f"""\
1 validation error for model-fields
field_b
  Field required [type=missing, input_value={{'field_a': b'abc'}}, input_type=dict]
    For further information visit https://errors.pydantic.dev/{pydantic_version}/v/missing"""
    )


@pytest.mark.parametrize(
    'config,input_value,expected',
    [
        ({}, {'a': '123'}, ({'a': 123, 'b': 4.2}, None, {'a'})),
        ({}, Map(a=123), ({'a': 123, 'b': 4.2}, None, {'a'})),
        ({}, {b'a': '123'}, Err('Field required [type=missing,')),
        ({}, {'a': '123', 'c': 4}, ({'a': 123, 'b': 4.2}, None, {'a'})),
        ({'extra_fields_behavior': 'allow'}, {'a': '123', 'c': 4}, ({'a': 123, 'b': 4.2}, {'c': 4}, {'a', 'c'})),
        ({'extra_fields_behavior': 'allow'}, {'a': '123', b'c': 4}, Err('Keys should be strings [type=invalid_key,')),
        (
            {'strict': True},
            Map(a=123),
            Err('Input should be a valid dictionary or instance of Model [type=model_type,'),
        ),
        ({}, {'a': '123', 'b': '4.7'}, ({'a': 123, 'b': 4.7}, None, {'a', 'b'})),
        ({}, {'a': '123', 'b': 'nan'}, ({'a': 123, 'b': FunctionCheck(math.isnan)}, None, {'a', 'b'})),
        (
            {'allow_inf_nan': False},
            {'a': '123', 'b': 'nan'},
            Err('Input should be a finite number [type=finite_number,'),
        ),
    ],
    ids=repr,
)
def test_config(config: CoreConfig, input_value, expected):
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'a': {'type': 'model-field', 'schema': {'type': 'int'}},
                'b': {
                    'type': 'model-field',
                    'schema': {'type': 'default', 'schema': {'type': 'float'}, 'default': 4.2},
                },
            },
        },
        config,
    )
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            val = v.validate_python(input_value)
            print(f'UNEXPECTED OUTPUT: {val!r}')
    else:
        result = v.validate_python(input_value)
        assert result == expected


def test_ignore_extra():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {'type': 'model-field', 'schema': {'type': 'str'}},
                'field_b': {'type': 'model-field', 'schema': {'type': 'int'}},
            },
        }
    )

    assert v.validate_python({'field_a': b'123', 'field_b': 1, 'field_c': 123}) == (
        {'field_a': '123', 'field_b': 1},
        None,
        {'field_b', 'field_a'},
    )


def test_forbid_extra():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'field_a': {'type': 'model-field', 'schema': {'type': 'str'}}},
            'extra_behavior': 'forbid',
        }
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'field_a': 'abc', 'field_b': 1})

    assert exc_info.value.errors(include_url=False) == [
        {'type': 'extra_forbidden', 'loc': ('field_b',), 'msg': 'Extra inputs are not permitted', 'input': 1}
    ]


def test_allow_extra_invalid():
    with pytest.raises(SchemaError, match='extras_schema can only be used if extra_behavior=allow'):
        SchemaValidator(
            {'type': 'model-fields', 'fields': {}, 'extras_schema': {'type': 'int'}, 'extra_behavior': 'ignore'}
        )


def test_allow_extra_wrong():
    with pytest.raises(SchemaError, match='Invalid extra_behavior: `wrong`'):
        SchemaValidator({'type': 'model-fields', 'fields': {}}, {'extra_fields_behavior': 'wrong'})


def test_str_config():
    v = SchemaValidator(
        {'type': 'model-fields', 'fields': {'field_a': {'type': 'model-field', 'schema': {'type': 'str'}}}},
        {'str_max_length': 5},
    )
    assert v.validate_python({'field_a': 'test'}) == ({'field_a': 'test'}, None, {'field_a'})

    with pytest.raises(ValidationError, match='String should have at most 5 characters'):
        v.validate_python({'field_a': 'test long'})


def test_validate_assignment():
    v = SchemaValidator(
        {'type': 'model-fields', 'fields': {'field_a': {'type': 'model-field', 'schema': {'type': 'str'}}}}
    )

    assert v.validate_python({'field_a': 'test'}) == ({'field_a': 'test'}, None, {'field_a'})

    data = {'field_a': 'test'}
    assert v.validate_assignment(data, 'field_a', b'abc') == ({'field_a': 'abc'}, None, {'field_a'})
    assert data == {'field_a': 'abc'}


def test_validate_assignment_strict_field():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'field_a': {'type': 'model-field', 'schema': {'type': 'str', 'strict': True}}},
        }
    )

    assert v.validate_python({'field_a': 'test'}) == ({'field_a': 'test'}, None, {'field_a'})

    with pytest.raises(ValidationError) as exc_info:
        v.validate_assignment({'field_a': 'test'}, 'field_a', b'abc')
    assert exc_info.value.errors(include_url=False) == [
        {'input': b'abc', 'type': 'string_type', 'loc': ('field_a',), 'msg': 'Input should be a valid string'}
    ]


def test_validate_assignment_functions():
    calls: list[Any] = []

    def func_a(input_value, info):
        calls.append(('func_a', input_value))
        return input_value * 2

    def func_b(input_value, info):
        calls.append(('func_b', input_value))
        return input_value / 2

    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {
                    'type': 'model-field',
                    'schema': {
                        'type': 'function-after',
                        'function': {'type': 'with-info', 'function': func_a},
                        'schema': {'type': 'str'},
                    },
                },
                'field_b': {
                    'type': 'model-field',
                    'schema': {
                        'type': 'function-after',
                        'function': {'type': 'with-info', 'function': func_b},
                        'schema': {'type': 'int'},
                    },
                },
            },
        }
    )

    assert v.validate_python({'field_a': 'test', 'field_b': 12.0}) == (
        {'field_a': 'testtest', 'field_b': 6},
        None,
        {'field_a', 'field_b'},
    )

    assert calls == [('func_a', 'test'), ('func_b', 12)]
    calls.clear()

    assert v.validate_assignment({'field_a': 'testtest', 'field_b': 6}, 'field_a', 'new-val') == (
        {'field_a': 'new-valnew-val', 'field_b': 6},
        None,
        {'field_a'},
    )
    assert calls == [('func_a', 'new-val')]


def test_validate_assignment_ignore_extra():
    v = SchemaValidator(
        {'type': 'model-fields', 'fields': {'field_a': {'type': 'model-field', 'schema': {'type': 'str'}}}}
    )

    assert v.validate_python({'field_a': 'test'}) == ({'field_a': 'test'}, None, {'field_a'})

    with pytest.raises(ValidationError) as exc_info:
        v.validate_assignment({'field_a': 'test'}, 'other_field', 456)

    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'no_such_attribute',
            'loc': ('other_field',),
            'msg': "Object has no attribute 'other_field'",
            'input': 456,
            'ctx': {'attribute': 'other_field'},
        }
    ]


def test_validate_assignment_allow_extra():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'field_a': {'type': 'model-field', 'schema': {'type': 'str'}}},
            'extra_behavior': 'allow',
        }
    )

    assert v.validate_python({'field_a': 'test'}) == ({'field_a': 'test'}, {}, {'field_a'})

    assert v.validate_assignment({'field_a': 'test'}, 'other_field', 456) == (
        {'field_a': 'test'},
        {'other_field': 456},
        {'other_field'},
    )


def test_validate_assignment_allow_extra_validate():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'field_a': {'type': 'model-field', 'schema': {'type': 'str'}}},
            'extras_schema': {'type': 'int'},
            'extra_behavior': 'allow',
        }
    )

    assert v.validate_assignment({'field_a': 'test'}, 'other_field', '456') == (
        {'field_a': 'test'},
        {'other_field': 456},
        {'other_field'},
    )

    with pytest.raises(ValidationError) as exc_info:
        assert v.validate_assignment({'field_a': 'test'}, 'other_field', 'xyz')
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': ('other_field',),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'xyz',
        }
    ]


def test_validate_assignment_with_strict():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'x': {'type': 'model-field', 'schema': {'type': 'str'}},
                'y': {'type': 'model-field', 'schema': {'type': 'int'}},
            },
        }
    )

    r, model_extra, fields_set = v.validate_python({'x': 'a', 'y': '123'})
    assert r == {'x': 'a', 'y': 123}
    assert model_extra is None
    assert fields_set == {'x', 'y'}

    v.validate_assignment(r, 'y', '124')
    assert r == {'x': 'a', 'y': 124}

    with pytest.raises(ValidationError) as exc_info:
        v.validate_assignment(r, 'y', '124', strict=True)

    assert exc_info.value.errors(include_url=False) == [
        {'type': 'int_type', 'loc': ('y',), 'msg': 'Input should be a valid integer', 'input': '124'}
    ]


def test_json_error():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'field_a': {'type': 'model-field', 'schema': {'type': 'list', 'items_schema': {'type': 'int'}}}},
        }
    )
    with pytest.raises(ValidationError) as exc_info:
        v.validate_json('{"field_a": [123, "wrong"]}')

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': ('field_a', 1),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'wrong',
        }
    ]


def test_missing_schema_key():
    with pytest.raises(SchemaError, match='model-fields.fields.x.schema\n  Field required'):
        validate_core_schema({'type': 'model-fields', 'fields': {'x': {'type': 'str'}}})


def test_fields_required_by_default():
    """By default all fields should be required"""
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'x': {'type': 'model-field', 'schema': {'type': 'str'}},
                'y': {'type': 'model-field', 'schema': {'type': 'str'}},
            },
        }
    )

    assert v.validate_python({'x': 'pika', 'y': 'chu'}) == ({'x': 'pika', 'y': 'chu'}, None, {'x', 'y'})

    with pytest.raises(ValidationError) as exc_info:
        assert v.validate_python({'x': 'pika'})

    assert exc_info.value.errors(include_url=False) == [
        {'type': 'missing', 'loc': ('y',), 'msg': 'Field required', 'input': {'x': 'pika'}}
    ]


def test_fields_required_by_default_with_default():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'x': {'type': 'model-field', 'schema': {'type': 'str'}},
                'y': {
                    'type': 'model-field',
                    'schema': {'type': 'default', 'schema': {'type': 'str'}, 'default': 'bulbi'},
                },
            },
        }
    )

    assert v.validate_python({'x': 'pika', 'y': 'chu'}) == ({'x': 'pika', 'y': 'chu'}, None, {'x', 'y'})
    assert v.validate_python({'x': 'pika'}) == ({'x': 'pika', 'y': 'bulbi'}, None, {'x'})


def test_alias(py_and_json: PyAndJson):
    v = py_and_json(
        {
            'type': 'model-fields',
            'fields': {'field_a': {'validation_alias': 'FieldA', 'type': 'model-field', 'schema': {'type': 'int'}}},
        }
    )
    assert v.validate_test({'FieldA': '123'}) == ({'field_a': 123}, None, {'field_a'})
    with pytest.raises(ValidationError, match=r'FieldA\n +Field required \[type=missing,'):
        assert v.validate_test({'foobar': '123'})
    with pytest.raises(ValidationError, match=r'FieldA\n +Field required \[type=missing,'):
        assert v.validate_test({'field_a': '123'})


def test_empty_string_field_name(py_and_json: PyAndJson):
    v = py_and_json({'type': 'model-fields', 'fields': {'': {'type': 'model-field', 'schema': {'type': 'int'}}}})
    assert v.validate_test({'': 123}) == ({'': 123}, None, {''})


def test_empty_string_aliases(py_and_json: PyAndJson):
    v = py_and_json(
        {
            'type': 'model-fields',
            'fields': {'field_a': {'validation_alias': '', 'type': 'model-field', 'schema': {'type': 'int'}}},
        }
    )
    assert v.validate_test({'': 123}) == ({'field_a': 123}, None, {'field_a'})

    v = py_and_json(
        {
            'type': 'model-fields',
            'fields': {'field_a': {'validation_alias': ['', ''], 'type': 'model-field', 'schema': {'type': 'int'}}},
        }
    )
    assert v.validate_test({'': {'': 123}}) == ({'field_a': 123}, None, {'field_a'})


def test_alias_allow_pop(py_and_json: PyAndJson):
    v = py_and_json(
        {
            'type': 'model-fields',
            'populate_by_name': True,
            'fields': {'field_a': {'validation_alias': 'FieldA', 'type': 'model-field', 'schema': {'type': 'int'}}},
        }
    )
    assert v.validate_test({'FieldA': '123'}) == ({'field_a': 123}, None, {'field_a'})
    assert v.validate_test({'field_a': '123'}) == ({'field_a': 123}, None, {'field_a'})
    assert v.validate_test({'FieldA': '1', 'field_a': '2'}) == ({'field_a': 1}, None, {'field_a'})
    with pytest.raises(ValidationError, match=r'FieldA\n +Field required \[type=missing,'):
        assert v.validate_test({'foobar': '123'})


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ({'foo': {'bar': '123'}}, ({'field_a': 123}, None, {'field_a'})),
        ({'x': '123'}, Err(r'foo.bar\n +Field required \[type=missing,')),
        ({'foo': '123'}, Err(r'foo.bar\n +Field required \[type=missing,')),
        ({'foo': [1, 2, 3]}, Err(r'foo.bar\n +Field required \[type=missing,')),
        ({'foo': {'bat': '123'}}, Err(r'foo.bar\n +Field required \[type=missing,')),
    ],
    ids=repr,
)
def test_alias_path(py_and_json: PyAndJson, input_value, expected):
    v = py_and_json(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {'validation_alias': ['foo', 'bar'], 'type': 'model-field', 'schema': {'type': 'int'}}
            },
        }
    )
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=expected.message):
            v.validate_test(input_value)
    else:
        output = v.validate_test(input_value)
        assert output == expected


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ({'foo': {'bar': {'bat': '123'}}}, ({'field_a': 123}, None, {'field_a'})),
        ({'foo': [1, 2, 3, 4]}, ({'field_a': 4}, None, {'field_a'})),
        ({'foo': (1, 2, 3, 4)}, ({'field_a': 4}, None, {'field_a'})),
        ({'spam': 5}, ({'field_a': 5}, None, {'field_a'})),
        ({'spam': 1, 'foo': {'bar': {'bat': 2}}}, ({'field_a': 2}, None, {'field_a'})),
        ({'foo': {'x': 2}}, Err(r'field_a\n +Field required \[type=missing,')),
        ({'x': '123'}, Err(r'field_a\n +Field required \[type=missing,')),
        ({'x': {2: 33}}, Err(r'field_a\n +Field required \[type=missing,')),
        ({'foo': '01234'}, Err(r'field_a\n +Field required \[type=missing,')),
        ({'foo': [1]}, Err(r'field_a\n +Field required \[type=missing,')),
    ],
    ids=repr,
)
def test_aliases_path_multiple(py_and_json: PyAndJson, input_value, expected):
    v = py_and_json(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {
                    'validation_alias': [['foo', 'bar', 'bat'], ['foo', 3], ['spam']],
                    'type': 'model-field',
                    'schema': {'type': 'int'},
                }
            },
        },
        {'loc_by_alias': False},
    )
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=expected.message):
            val = v.validate_test(input_value)
            print(f'UNEXPECTED OUTPUT: {val!r}')
    else:
        output = v.validate_test(input_value)
        assert output == expected


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ({'foo': {-2: '123'}}, ({'field_a': 123}, None, {'field_a'})),
        # negatives indexes work fine
        ({'foo': [1, 42, 'xx']}, ({'field_a': 42}, None, {'field_a'})),
        ({'foo': [42, 'xxx', 42]}, Err(r'Input should be a valid integer,')),
        ({'foo': [42]}, Err(r'field_a\n +Field required \[type=missing,')),
        ({'foo': {'xx': '123'}}, Err(r'field_a\n +Field required \[type=missing,')),
        ({'foo': {'-2': '123'}}, Err(r'field_a\n +Field required \[type=missing,')),
        ({'foo': {2: '123'}}, Err(r'field_a\n +Field required \[type=missing,')),
        ({'foo': 'foobar'}, Err(r'field_a\n +Field required \[type=missing,')),
        ({'foo': {0, 1, 2}}, Err(r'field_a\n +Field required \[type=missing,')),
    ],
    ids=repr,
)
def test_aliases_path_negative(input_value, expected):
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'field_a': {'validation_alias': ['foo', -2], 'type': 'model-field', 'schema': {'type': 'int'}}},
        },
        {'loc_by_alias': False},
    )
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=expected.message):
            val = v.validate_python(input_value)
            print(f'UNEXPECTED OUTPUT: {val!r}')
    else:
        output = v.validate_python(input_value)
        assert output == expected


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ({'foo': [1, 42, 'xx']}, ({'field_a': 42}, None, {'field_a'})),
        ({'foo': [42, 'xxx', 42]}, Err(r'Input should be a valid integer,')),
        ({'foo': [42]}, Err(r'foo.-2\n +Field required \[type=missing,')),
    ],
    ids=repr,
)
def test_aliases_path_negative_json(py_and_json: PyAndJson, input_value, expected):
    v = py_and_json(
        {
            'type': 'model-fields',
            'fields': {'field_a': {'validation_alias': ['foo', -2], 'type': 'model-field', 'schema': {'type': 'int'}}},
        }
    )
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=expected.message):
            val = v.validate_test(input_value)
            print(f'UNEXPECTED OUTPUT: {val!r}')
    else:
        output = v.validate_test(input_value)
        assert output == expected


def test_aliases_debug():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {
                    'validation_alias': [['foo', 'bar', 'bat'], ['foo', 3]],
                    'type': 'model-field',
                    'schema': {'type': 'int'},
                }
            },
        }
    )
    print(repr(v))
    assert repr(v).startswith('SchemaValidator(title="model-fields", validator=ModelFields(')
    assert 'PathChoices(' in repr(v)


def get_int_key():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {
                    'validation_alias': [['foo', 3], ['spam']],
                    'type': 'model-field',
                    'schema': {'type': 'int'},
                }
            },
        }
    )
    assert v.validate_python({'foo': {3: 33}}) == ({'field_a': 33}, {}, {'field_a'})


class GetItemThing:
    def __getitem__(self, v):
        assert v == 'foo'
        return 321


def get_custom_getitem():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'field_a': {'validation_alias': ['foo'], 'type': 'model-field', 'schema': {'type': 'int'}}},
        }
    )
    assert v.validate_python(GetItemThing()) == ({'field_a': 321}, {}, {'field_a'})
    assert v.validate_python({'bar': GetItemThing()}) == ({'field_a': 321}, {}, {'field_a'})


@pytest.mark.parametrize('input_value', [{'foo': {'bar': 42}}, {'foo': 42}, {'field_a': 42}], ids=repr)
def test_paths_allow_by_name(py_and_json: PyAndJson, input_value):
    v = py_and_json(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {
                    'validation_alias': [['foo', 'bar'], ['foo']],
                    'type': 'model-field',
                    'schema': {'type': 'int'},
                }
            },
            'populate_by_name': True,
        }
    )
    assert v.validate_test(input_value) == ({'field_a': 42}, None, {'field_a'})


@pytest.mark.parametrize(
    'alias_schema,error',
    [
        ({'validation_alias': ['foo', ['bar']]}, 'Input should be a valid string'),
        ({'validation_alias': []}, 'Lookup paths should have at least one element'),
        ({'validation_alias': [[]]}, 'Each alias path should have at least one element'),
        ({'validation_alias': [123]}, "TypeError: 'int' object cannot be converted to 'PyList'"),
        ({'validation_alias': [[[]]]}, 'Input should be a valid string'),
        ({'validation_alias': [[1, 'foo']]}, 'TypeError: The first item in an alias path should be a string'),
    ],
    ids=repr,
)
def test_alias_build_error(alias_schema, error):
    with pytest.raises(SchemaError, match=error):
        SchemaValidator(
            validate_core_schema(
                {
                    'type': 'model-fields',
                    'fields': {'field_a': {'type': 'model-field', 'schema': {'type': 'int'}, **alias_schema}},
                }
            )
        )


def test_alias_error_loc_alias(py_and_json: PyAndJson):
    v = py_and_json(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {
                    'type': 'model-field',
                    'schema': {'type': 'int'},
                    'validation_alias': [['foo', 'x'], ['bar', 1, -1]],
                }
            },
        },
        {'loc_by_alias': True},  # this is the default
    )
    assert v.validate_test({'foo': {'x': 42}}) == ({'field_a': 42}, None, {'field_a'})
    assert v.validate_python({'bar': ['x', {-1: 42}]}) == ({'field_a': 42}, None, {'field_a'})
    assert v.validate_test({'bar': ['x', [1, 2, 42]]}) == ({'field_a': 42}, None, {'field_a'})
    with pytest.raises(ValidationError) as exc_info:
        v.validate_test({'foo': {'x': 'not_int'}})
    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': ('foo', 'x'),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'not_int',
        }
    ]
    with pytest.raises(ValidationError) as exc_info:
        v.validate_test({'bar': ['x', [1, 2, 'not_int']]})
    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': ('bar', 1, -1),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'not_int',
        }
    ]
    with pytest.raises(ValidationError) as exc_info:
        v.validate_test({})
    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.errors(include_url=False) == [
        {'type': 'missing', 'loc': ('foo', 'x'), 'msg': 'Field required', 'input': {}}
    ]


def test_alias_error_loc_field_names(py_and_json: PyAndJson):
    v = py_and_json(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {
                    'type': 'model-field',
                    'schema': {'type': 'int'},
                    'validation_alias': [['foo'], ['bar', 1, -1]],
                }
            },
        },
        {'loc_by_alias': False},
    )
    assert v.validate_test({'foo': 42}) == ({'field_a': 42}, None, {'field_a'})
    assert v.validate_test({'bar': ['x', [1, 2, 42]]}) == ({'field_a': 42}, None, {'field_a'})
    with pytest.raises(ValidationError) as exc_info:
        v.validate_test({'foo': 'not_int'})
    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': ('field_a',),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'not_int',
        }
    ]
    with pytest.raises(ValidationError) as exc_info:
        v.validate_test({'bar': ['x', [1, 2, 'not_int']]})
    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': ('field_a',),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'not_int',
        }
    ]
    with pytest.raises(ValidationError) as exc_info:
        v.validate_test({})
    # insert_assert(exc_info.value.errors(include_url=False))
    assert exc_info.value.errors(include_url=False) == [
        {'type': 'missing', 'loc': ('field_a',), 'msg': 'Field required', 'input': {}}
    ]


def test_empty_model():
    v = SchemaValidator({'type': 'model-fields', 'fields': {}})
    assert v.validate_python({}) == ({}, None, set())
    with pytest.raises(
        ValidationError, match=re.escape('Input should be a valid dictionary or instance of Model [type=model_type,')
    ):
        v.validate_python('x')


def test_model_fields_deep():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'field_a': {'type': 'model-field', 'schema': {'type': 'str'}},
                'field_b': {
                    'type': 'model-field',
                    'schema': {
                        'type': 'model-fields',
                        'fields': {
                            'field_c': {'type': 'model-field', 'schema': {'type': 'str'}},
                            'field_d': {
                                'type': 'model-field',
                                'schema': {
                                    'type': 'model-fields',
                                    'fields': {
                                        'field_e': {'type': 'model-field', 'schema': {'type': 'str'}},
                                        'field_f': {'type': 'model-field', 'schema': {'type': 'int'}},
                                    },
                                },
                            },
                        },
                    },
                },
            },
        }
    )
    model_dict, model_extra, fields_set = v.validate_python(
        {'field_a': '1', 'field_b': {'field_c': '2', 'field_d': {'field_e': '4', 'field_f': 4}}}
    )
    assert model_dict == {
        'field_a': '1',
        'field_b': (
            {'field_c': '2', 'field_d': ({'field_e': '4', 'field_f': 4}, None, {'field_f', 'field_e'})},
            None,
            {'field_d', 'field_c'},
        ),
    }
    assert model_extra is None
    assert fields_set == {'field_a', 'field_b'}
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'field_a': '1', 'field_b': {'field_c': '2', 'field_d': {'field_e': '4', 'field_f': 'xx'}}})

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': ('field_b', 'field_d', 'field_f'),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': 'xx',
        }
    ]


class ClassWithAttributes:
    def __init__(self):
        self.a = 1
        self.b = 2

    @property
    def c(self):
        return 'ham'


@dataclass
class MyDataclass:
    a: int = 1
    b: int = 2
    c: str = 'ham'


@pytest.mark.parametrize(
    'input_value,expected',
    [
        (ClassWithAttributes(), ({'a': 1, 'b': 2, 'c': 'ham'}, None, {'a', 'b', 'c'})),
        (MyDataclass(), ({'a': 1, 'b': 2, 'c': 'ham'}, None, {'a', 'b', 'c'})),
        (Cls(a=1, b=2, c='ham'), ({'a': 1, 'b': 2, 'c': 'ham'}, None, {'a', 'b', 'c'})),
        (dict(a=1, b=2, c='ham'), ({'a': 1, 'b': 2, 'c': 'ham'}, None, {'a', 'b', 'c'})),
        (Map(a=1, b=2, c='ham'), ({'a': 1, 'b': 2, 'c': 'ham'}, None, {'a', 'b', 'c'})),
        ((Cls(a=1, b=2), dict(c='ham')), ({'a': 1, 'b': 2, 'c': 'ham'}, None, {'a', 'b', 'c'})),
        ((Cls(a=1, b=2), dict(c='bacon')), ({'a': 1, 'b': 2, 'c': 'bacon'}, None, {'a', 'b', 'c'})),
        ((Cls(a=1, b=2, c='ham'), dict(c='bacon')), ({'a': 1, 'b': 2, 'c': 'bacon'}, None, {'a', 'b', 'c'})),
        ((Cls(a=1, b=2, c='ham'), dict(d='bacon')), ({'a': 1, 'b': 2, 'c': 'ham'}, None, {'a', 'b', 'c'})),
        # using type gives `__module__ == 'builtins'`
        (type('Testing', (), {}), Err('[type=model_attributes_type,')),
        (
            '123',
            Err('Input should be a valid dictionary or object to extract fields from [type=model_attributes_type,'),
        ),
        ([(1, 2)], Err('type=model_attributes_type,')),
        (((1, 2),), Err('type=model_attributes_type,')),
    ],
    ids=repr,
)
@pytest.mark.parametrize('from_attributes_mode', ['schema', 'validation'])
def test_from_attributes(input_value, expected, from_attributes_mode):
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'a': {'type': 'model-field', 'schema': {'type': 'int'}},
                'b': {'type': 'model-field', 'schema': {'type': 'int'}},
                'c': {'type': 'model-field', 'schema': {'type': 'str'}},
            },
            'from_attributes': from_attributes_mode == 'schema',
        }
    )
    kwargs = {}
    if from_attributes_mode == 'validation':
        kwargs['from_attributes'] = True
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            val = v.validate_python(input_value, **kwargs)
            print(f'UNEXPECTED OUTPUT: {val!r}')
    else:
        output = v.validate_python(input_value, **kwargs)
        assert output == expected


def test_from_attributes_type_error():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'a': {'type': 'model-field', 'schema': {'type': 'int'}},
                'b': {'type': 'model-field', 'schema': {'type': 'int'}},
                'c': {'type': 'model-field', 'schema': {'type': 'str'}},
            },
            'from_attributes': True,
            'model_name': 'MyModel',
        }
    )
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python('123')

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'model_attributes_type',
            'loc': (),
            'msg': 'Input should be a valid dictionary or object to extract fields from',
            'input': '123',
        }
    ]

    with pytest.raises(ValidationError) as exc_info:
        v.validate_json('123')
    # insert_assert(exc_info.value.errors())
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'model_type',
            'loc': (),
            'msg': 'Input should be an object',
            'input': 123,
            'ctx': {'class_name': 'MyModel'},
        }
    ]


def test_from_attributes_by_name():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'a': {'type': 'model-field', 'schema': {'type': 'int'}, 'validation_alias': 'a_alias'}},
            'from_attributes': True,
            'populate_by_name': True,
        }
    )
    assert v.validate_python(Cls(a_alias=1)) == ({'a': 1}, None, {'a'})
    assert v.validate_python(Cls(a=1)) == ({'a': 1}, None, {'a'})


def test_from_attributes_override_true():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'a': {'type': 'model-field', 'schema': {'type': 'int'}}},
            'from_attributes': False,
        }
    )
    with pytest.raises(ValidationError, match='Input should be a valid dictionary'):
        v.validate_python(Cls(a=1))
    assert v.validate_python(Cls(a=1), from_attributes=True) == ({'a': 1}, None, {'a'})

    assert v.isinstance_python(Cls(a=1), from_attributes=True) is True
    assert v.isinstance_python(Cls(a=1)) is False


def test_from_attributes_override_false():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'a': {'type': 'model-field', 'schema': {'type': 'int'}}},
            'from_attributes': True,
        }
    )
    with pytest.raises(ValidationError, match='Input should be a valid dictionary'):
        v.validate_python(Cls(a=1), from_attributes=False)
    assert v.validate_python(Cls(a=1)) == ({'a': 1}, None, {'a'})

    assert v.isinstance_python(Cls(a=1)) is True
    assert v.isinstance_python(Cls(a=1), from_attributes=False) is False


def test_from_attributes_missing():
    class Foobar:
        def __init__(self):
            self.a = 1
            self.b = 2

    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'a': {'type': 'model-field', 'schema': {'type': 'int'}},
                'b': {'type': 'model-field', 'schema': {'type': 'int'}},
                'c': {'type': 'model-field', 'schema': {'type': 'str'}},
            },
            'from_attributes': True,
        }
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(Foobar())

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'missing',
            'loc': ('c',),
            'msg': 'Field required',
            'input': HasRepr(IsStr(regex='.+Foobar object at.+')),
        }
    ]


def test_from_attributes_error():
    class Foobar:
        def __init__(self):
            self.a = 1

        @property
        def b(self):
            raise RuntimeError('intentional error')

    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'a': {'type': 'model-field', 'schema': {'type': 'int'}},
                'b': {'type': 'model-field', 'schema': {'type': 'int'}},
            },
            'from_attributes': True,
        }
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(Foobar())

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'get_attribute_error',
            'loc': ('b',),
            'msg': 'Error extracting attribute: RuntimeError: intentional error',
            'input': HasRepr(IsStr(regex='.+Foobar object at.+')),
            'ctx': {'error': 'RuntimeError: intentional error'},
        }
    ]


def test_from_attributes_extra():
    def another_function(x):
        return x

    class Foobar:
        def __init__(self):
            self.a = 1
            self.b = 2
            self._private_attribute = 4

        @property
        def c(self):
            return 'ham'

        @property
        def _private_property(self):
            return 'wrong'

        @property
        def property_error(self):
            raise RuntimeError('xxx')

        def bound_method(self):
            return f'wrong {self.a}'

        @staticmethod
        def static_method():
            return 'wrong'

        # this is omitted along with the static method by the !PyFunction::is_type_of(attr) check in fields
        function_attribute = another_function

        @classmethod
        def class_method(cls):
            return 'wrong'

    @dataclass
    class MyDataclass:
        a: int = 1
        b: int = 2
        c: str = 'ham'
        _d: int = 4

    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'a': {'type': 'model-field', 'schema': {'type': 'int'}}},
            'from_attributes': True,
            'extra_behavior': 'allow',
        }
    )

    assert v.validate_python(Foobar()) == ({'a': 1}, {}, {'a'})
    assert v.validate_python(MyDataclass()) == ({'a': 1}, {}, {'a'})
    assert v.validate_python(Cls(a=1, b=2, c='ham')) == ({'a': 1}, {}, {'a'})
    assert v.validate_python(Cls(a=1, b=datetime(2000, 1, 1))) == ({'a': 1}, {}, {'a'})
    assert v.validate_python(Cls(a=1, b=datetime.now, c=lambda: 42)) == ({'a': 1}, {}, {'a'})


def test_from_attributes_extra_ignore_no_attributes_accessed() -> None:
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'a': {'type': 'model-field', 'schema': {'type': 'int'}}},
            'from_attributes': True,
            'extra_behavior': 'ignore',
        }
    )

    accessed: List[str] = []

    class Source:
        a = 1
        b = 2

        def __getattribute__(self, name: str, /) -> Any:
            accessed.append(name)
            return super().__getattribute__(name)

    assert v.validate_python(Source()) == ({'a': 1}, None, {'a'})
    assert 'a' in accessed and 'b' not in accessed


def test_from_attributes_extra_forbid() -> None:
    class Source:
        a = 1
        b = 2

    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'a': {'type': 'model-field', 'schema': {'type': 'int'}}},
            'from_attributes': True,
            'extra_behavior': 'forbid',
        }
    )

    assert v.validate_python(Source()) == ({'a': 1}, None, {'a'})


def foobar():
    pass


@pytest.mark.parametrize(
    'input_value,expected',
    [
        (Cls(a=1), {'a': 1}),
        (Cls(a=datetime.now), {'a': datetime.now}),
        (Cls(a=lambda: 42), {'a': HasRepr(IsStr(regex='.+<lambda>.+'))}),
        (Cls(a=sys.path), {'a': sys.path}),
        (Cls(a=foobar), {'a': foobar}),
    ],
    ids=repr,
)
def test_from_attributes_function(input_value, expected):
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'a': {'type': 'model-field', 'schema': {'type': 'any'}}},
            'from_attributes': True,
        }
    )

    model_dict, model_extra, fields_set = v.validate_python(input_value)
    assert model_dict == expected
    assert model_extra is None
    assert fields_set == {'a'}


def test_from_attributes_error_error():
    class BadError(Exception):
        def __str__(self):
            raise RuntimeError('intentional error inside error')

    class Foobar:
        @property
        def x(self):
            raise BadError('intentional error')

    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {'x': {'type': 'model-field', 'schema': {'type': 'int'}}},
            'from_attributes': True,
        }
    )

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(Foobar())

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'get_attribute_error',
            'loc': ('x',),
            'msg': IsStr(regex=r'Error extracting attribute: \S+\.<locals>\.BadError: <exception str\(\) failed>'),
            'input': HasRepr(IsStr(regex='.+Foobar object at.+')),
            'ctx': {'error': IsStr(regex=r'\S+\.<locals>\.BadError: <exception str\(\) failed>')},
        }
    ]

    class UnInitError:
        @property
        def x(self):
            raise RuntimeError

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(UnInitError())

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'get_attribute_error',
            'loc': ('x',),
            'msg': 'Error extracting attribute: RuntimeError',
            'input': HasRepr(IsStr(regex='.+UnInitError object at.+')),
            'ctx': {'error': 'RuntimeError'},
        }
    ]


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ({'foo': {'bar': {'bat': '123'}}}, {'my_field': 123}),
        (Cls(foo=Cls(bar=Cls(bat='123'))), {'my_field': 123}),
        (Cls(foo={'bar': {'bat': '123'}}), {'my_field': 123}),
        (Cls(foo=[1, 2, 3, 4]), {'my_field': 4}),
        (Cls(foo=(1, 2, 3, 4)), {'my_field': 4}),
        (Cls(spam=5), {'my_field': 5}),
        (Cls(spam=1, foo=Cls(bar=Cls(bat=2))), {'my_field': 2}),
        (Cls(x='123'), Err(r'my_field\n +Field required \[type=missing,')),
        (Cls(x={2: 33}), Err(r'my_field\n +Field required \[type=missing,')),
        (Cls(foo='01234'), Err(r'my_field\n +Field required \[type=missing,')),
        (Cls(foo=[1]), Err(r'my_field\n +Field required \[type=missing,')),
        (Cls, Err(r'Input should be a valid dictionary')),
    ],
    ids=repr,
)
def test_from_attributes_path(input_value, expected):
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'my_field': {
                    'validation_alias': [['foo', 'bar', 'bat'], ['foo', 3], ['spam']],
                    'type': 'model-field',
                    'schema': {'type': 'int'},
                }
            },
            'from_attributes': True,
        },
        {'loc_by_alias': False},
    )
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=expected.message):
            val = v.validate_python(input_value)
            print(f'UNEXPECTED OUTPUT: {val!r}')
    else:
        model_dict, model_extra, fields_set = v.validate_python(input_value)
        assert model_dict == expected
        assert model_extra is None
        assert fields_set == {'my_field'}


def test_from_attributes_path_error():
    class PropertyError:
        @property
        def foo(self):
            raise RuntimeError('intentional error')

    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'my_field': {
                    'validation_alias': [['foo', 'bar', 'bat'], ['foo', 3], ['spam']],
                    'type': 'model-field',
                    'schema': {'type': 'int'},
                }
            },
            'from_attributes': True,
        }
    )
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(PropertyError())

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'get_attribute_error',
            'loc': ('my_field',),
            'msg': 'Error extracting attribute: RuntimeError: intentional error',
            'input': HasRepr(IsStr(regex='.+PropertyError object at.+')),
            'ctx': {'error': 'RuntimeError: intentional error'},
        }
    ]


def test_alias_extra(py_and_json: PyAndJson):
    v = py_and_json(
        {
            'type': 'model-fields',
            'extra_behavior': 'allow',
            'fields': {
                'field_a': {
                    'validation_alias': [['FieldA'], ['foo', 2]],
                    'type': 'model-field',
                    'schema': {'type': 'int'},
                }
            },
        },
        {'loc_by_alias': False},
    )
    assert v.validate_test({'FieldA': 1}) == ({'field_a': 1}, {}, {'field_a'})
    assert v.validate_test({'foo': [1, 2, 3]}) == ({'field_a': 3}, {}, {'field_a'})

    # used_keys should be populated either though validation fails so "FieldA" is skipped in extra
    with pytest.raises(ValidationError) as exc_info:
        assert v.validate_test({'FieldA': '...'}) == ({'field_a': 1}, {}, {'field_a'})

    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'int_parsing',
            'loc': ('field_a',),
            'msg': 'Input should be a valid integer, unable to parse string as an integer',
            'input': '...',
        }
    ]


def test_alias_extra_from_attributes():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'extra_behavior': 'allow',
            'from_attributes': True,
            'fields': {
                'field_a': {
                    'validation_alias': [['FieldA'], ['foo', 2]],
                    'type': 'model-field',
                    'schema': {'type': 'int'},
                }
            },
        }
    )
    assert v.validate_python({'FieldA': 1}) == ({'field_a': 1}, {}, {'field_a'})
    assert v.validate_python(Cls(FieldA=1)) == ({'field_a': 1}, {}, {'field_a'})
    assert v.validate_python(Cls(foo=[1, 2, 3])) == ({'field_a': 3}, {}, {'field_a'})
    assert v.validate_python({'foo': [1, 2, 3]}) == ({'field_a': 3}, {}, {'field_a'})


def test_alias_extra_by_name(py_and_json: PyAndJson):
    v = py_and_json(
        {
            'type': 'model-fields',
            'extra_behavior': 'allow',
            'from_attributes': True,
            'populate_by_name': True,
            'fields': {'field_a': {'validation_alias': 'FieldA', 'type': 'model-field', 'schema': {'type': 'int'}}},
        }
    )
    assert v.validate_test({'FieldA': 1}) == ({'field_a': 1}, {}, {'field_a'})
    assert v.validate_test({'field_a': 1}) == ({'field_a': 1}, {}, {'field_a'})
    assert v.validate_python(Cls(FieldA=1)) == ({'field_a': 1}, {}, {'field_a'})
    assert v.validate_python(Cls(field_a=1)) == ({'field_a': 1}, {}, {'field_a'})


def test_alias_extra_forbid(py_and_json: PyAndJson):
    v = py_and_json(
        {
            'type': 'model-fields',
            'extra_behavior': 'forbid',
            'fields': {'field_a': {'type': 'model-field', 'validation_alias': 'FieldA', 'schema': {'type': 'int'}}},
        }
    )
    assert v.validate_test({'FieldA': 1}) == ({'field_a': 1}, None, {'field_a'})


def test_with_default_factory():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'x': {
                    'type': 'model-field',
                    'schema': {'type': 'default', 'schema': {'type': 'str'}, 'default_factory': lambda: 'pikachu'},
                }
            },
        }
    )

    assert v.validate_python({}) == ({'x': 'pikachu'}, None, set())
    assert v.validate_python({'x': 'bulbi'}) == ({'x': 'bulbi'}, None, {'x'})


@pytest.mark.parametrize(
    'default_factory,error_message',
    [
        (lambda: 1 + 'a', "unsupported operand type(s) for +: 'int' and 'str'"),
        (lambda x: 'a' + x, "<lambda>() missing 1 required positional argument: 'x'"),
    ],
)
def test_bad_default_factory(default_factory, error_message):
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'x': {
                    'type': 'model-field',
                    'schema': {'type': 'default', 'schema': {'type': 'str'}, 'default_factory': default_factory},
                }
            },
        }
    )
    with pytest.raises(TypeError, match=re.escape(error_message)):
        v.validate_python({})


class TestOnError:
    def test_on_error_bad_name(self):
        with pytest.raises(SchemaError, match="Input should be 'raise', 'omit' or 'default'"):
            validate_core_schema(
                {
                    'type': 'model-fields',
                    'fields': {
                        'x': {
                            'type': 'model-field',
                            'schema': {'type': 'default', 'schema': {'type': 'str'}, 'on_error': 'rais'},
                        }
                    },
                }
            )

    def test_on_error_bad_default(self):
        with pytest.raises(SchemaError, match="'on_error = default' requires a `default` or `default_factory`"):
            SchemaValidator(
                {
                    'type': 'model-fields',
                    'fields': {
                        'x': {
                            'type': 'model-field',
                            'schema': {'type': 'default', 'schema': {'type': 'str'}, 'on_error': 'default'},
                        }
                    },
                }
            )

    def test_on_error_raise_by_default(self, py_and_json: PyAndJson):
        v = py_and_json({'type': 'model-fields', 'fields': {'x': {'type': 'model-field', 'schema': {'type': 'str'}}}})
        assert v.validate_test({'x': 'foo'}) == ({'x': 'foo'}, None, {'x'})
        with pytest.raises(ValidationError) as exc_info:
            v.validate_test({'x': ['foo']})
        assert exc_info.value.errors(include_url=False) == [
            {'input': ['foo'], 'type': 'string_type', 'loc': ('x',), 'msg': 'Input should be a valid string'}
        ]

    def test_on_error_raise_explicit(self, py_and_json: PyAndJson):
        v = py_and_json(
            {
                'type': 'model-fields',
                'fields': {
                    'x': {
                        'type': 'model-field',
                        'schema': {'type': 'default', 'schema': {'type': 'str'}, 'on_error': 'raise'},
                    }
                },
            }
        )
        assert v.validate_test({'x': 'foo'}) == ({'x': 'foo'}, None, {'x'})
        with pytest.raises(ValidationError) as exc_info:
            v.validate_test({'x': ['foo']})
        assert exc_info.value.errors(include_url=False) == [
            {'input': ['foo'], 'type': 'string_type', 'loc': ('x',), 'msg': 'Input should be a valid string'}
        ]

    def test_on_error_default(self, py_and_json: PyAndJson):
        v = py_and_json(
            {
                'type': 'model-fields',
                'fields': {
                    'x': {
                        'type': 'model-field',
                        'schema': {
                            'type': 'default',
                            'schema': {'type': 'str'},
                            'on_error': 'default',
                            'default': 'pika',
                        },
                    }
                },
            }
        )
        assert v.validate_test({'x': 'foo'}) == ({'x': 'foo'}, None, {'x'})
        assert v.validate_test({'x': ['foo']}) == ({'x': 'pika'}, None, {'x'})

    def test_on_error_default_factory(self, py_and_json: PyAndJson):
        v = py_and_json(
            {
                'type': 'model-fields',
                'fields': {
                    'x': {
                        'type': 'model-field',
                        'schema': {
                            'type': 'default',
                            'schema': {'type': 'str'},
                            'on_error': 'default',
                            'default_factory': lambda: 'pika',
                        },
                    }
                },
            }
        )
        assert v.validate_test({'x': 'foo'}) == ({'x': 'foo'}, None, {'x'})
        assert v.validate_test({'x': ['foo']}) == ({'x': 'pika'}, None, {'x'})

    def test_wrap_on_error(self, py_and_json: PyAndJson):
        def wrap_function(input_value, validator, info):
            try:
                return validator(input_value)
            except ValidationError:
                if isinstance(input_value, list):
                    return str(len(input_value))
                else:
                    return repr(input_value)

        v = py_and_json(
            {
                'type': 'model-fields',
                'fields': {
                    'x': {
                        'type': 'model-field',
                        'schema': {
                            'type': 'default',
                            'on_error': 'raise',
                            'schema': {
                                'type': 'function-wrap',
                                'function': {'type': 'with-info', 'function': wrap_function},
                                'schema': {'type': 'str'},
                            },
                        },
                    }
                },
            }
        )
        assert v.validate_test({'x': 'foo'}) == ({'x': 'foo'}, None, {'x'})
        assert v.validate_test({'x': ['foo']}) == ({'x': '1'}, None, {'x'})
        assert v.validate_test({'x': ['foo', 'bar']}) == ({'x': '2'}, None, {'x'})
        assert v.validate_test({'x': {'a': 'b'}}) == ({'x': "{'a': 'b'}"}, None, {'x'})


def test_frozen_field():
    v = SchemaValidator(
        {
            'type': 'model-fields',
            'fields': {
                'name': {'type': 'model-field', 'schema': {'type': 'str'}},
                'age': {'type': 'model-field', 'schema': {'type': 'int'}},
                'is_developer': {
                    'type': 'model-field',
                    'schema': {'type': 'default', 'schema': {'type': 'bool'}, 'default': True},
                    'frozen': True,
                },
            },
        }
    )
    r1, model_extra, fields_set = v.validate_python({'name': 'Samuel', 'age': '36'})
    assert r1 == {'name': 'Samuel', 'age': 36, 'is_developer': True}
    assert model_extra is None
    assert fields_set == {'name', 'age'}
    v.validate_assignment(r1, 'age', '35')
    assert r1 == {'name': 'Samuel', 'age': 35, 'is_developer': True}
    with pytest.raises(ValidationError) as exc_info:
        v.validate_assignment(r1, 'is_developer', False)
    assert exc_info.value.errors(include_url=False) == [
        {'type': 'frozen_field', 'loc': ('is_developer',), 'msg': 'Field is frozen', 'input': False}
    ]


@pytest.mark.parametrize(
    'config,schema_extra_behavior_kw',
    [
        (core_schema.CoreConfig(extra_fields_behavior='allow'), {}),
        (core_schema.CoreConfig(extra_fields_behavior='allow'), {'extra_behavior': None}),
        (core_schema.CoreConfig(), {'extra_behavior': 'allow'}),
        (None, {'extra_behavior': 'allow'}),
        (core_schema.CoreConfig(extra_fields_behavior='forbid'), {'extra_behavior': 'allow'}),
    ],
)
@pytest.mark.parametrize(
    'extras_schema_kw, expected_extra_value',
    [({}, '123'), ({'extras_schema': None}, '123'), ({'extras_schema': core_schema.int_schema()}, 123)],
    ids=['extras_schema=unset', 'extras_schema=None', 'extras_schema=int'],
)
def test_extra_behavior_allow(
    config: Union[core_schema.CoreConfig, None],
    schema_extra_behavior_kw: Dict[str, Any],
    extras_schema_kw: Dict[str, Any],
    expected_extra_value: Any,
):
    v = SchemaValidator(
        core_schema.model_fields_schema(
            {'f': core_schema.model_field(core_schema.str_schema())}, **schema_extra_behavior_kw, **extras_schema_kw
        ),
        config=config,
    )

    m, model_extra, fields_set = v.validate_python({'f': 'x', 'extra_field': '123'})
    assert m == {'f': 'x'}
    assert model_extra == {'extra_field': expected_extra_value}
    assert fields_set == {'f', 'extra_field'}

    v.validate_assignment(m, 'f', 'y')
    assert m == {'f': 'y'}

    new_m, new_model_extra, new_fields_set = v.validate_assignment({**m, **model_extra}, 'not_f', '123')
    assert new_m == {'f': 'y'}
    assert new_model_extra == {'extra_field': expected_extra_value, 'not_f': expected_extra_value}
    assert new_fields_set == {'not_f'}


@pytest.mark.parametrize(
    'config,schema_extra_behavior_kw',
    [
        (core_schema.CoreConfig(extra_fields_behavior='forbid'), {}),
        (core_schema.CoreConfig(extra_fields_behavior='forbid'), {'extra_behavior': None}),
        (core_schema.CoreConfig(), {'extra_behavior': 'forbid'}),
        (None, {'extra_behavior': 'forbid'}),
        (core_schema.CoreConfig(extra_fields_behavior='allow'), {'extra_behavior': 'forbid'}),
    ],
)
def test_extra_behavior_forbid(config: Union[core_schema.CoreConfig, None], schema_extra_behavior_kw: Dict[str, Any]):
    v = SchemaValidator(
        core_schema.model_fields_schema(
            {'f': core_schema.model_field(core_schema.str_schema())}, **schema_extra_behavior_kw
        ),
        config=config,
    )

    m, model_extra, fields_set = v.validate_python({'f': 'x'})
    assert m == {'f': 'x'}
    assert fields_set == {'f'}

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'f': 'x', 'extra_field': 123})
    assert exc_info.value.errors(include_url=False) == [
        {'type': 'extra_forbidden', 'loc': ('extra_field',), 'msg': 'Extra inputs are not permitted', 'input': 123}
    ]

    v.validate_assignment(m, 'f', 'y')
    assert m['f'] == 'y'

    with pytest.raises(ValidationError) as exc_info:
        v.validate_assignment(m, 'not_f', 'xyz')
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'no_such_attribute',
            'loc': ('not_f',),
            'msg': "Object has no attribute 'not_f'",
            'input': 'xyz',
            'ctx': {'attribute': 'not_f'},
        }
    ]
    assert 'not_f' not in m


@pytest.mark.parametrize(
    'config,schema_extra_behavior_kw',
    [
        (core_schema.CoreConfig(extra_fields_behavior='ignore'), {}),
        (core_schema.CoreConfig(), {'extra_behavior': 'ignore'}),
        (None, {'extra_behavior': 'ignore'}),
        (core_schema.CoreConfig(extra_fields_behavior='forbid'), {'extra_behavior': 'ignore'}),
        (core_schema.CoreConfig(), {}),
        (core_schema.CoreConfig(), {'extra_behavior': None}),
        (None, {'extra_behavior': None}),
    ],
)
def test_extra_behavior_ignore(config: Union[core_schema.CoreConfig, None], schema_extra_behavior_kw: Dict[str, Any]):
    v = SchemaValidator(
        core_schema.model_fields_schema(
            {'f': core_schema.model_field(core_schema.str_schema())}, **schema_extra_behavior_kw
        ),
        config=config,
    )

    m, model_extra, fields_set = v.validate_python({'f': 'x', 'extra_field': 123})
    assert m == {'f': 'x'}
    assert model_extra is None
    assert fields_set == {'f'}

    v.validate_assignment(m, 'f', 'y')
    assert m['f'] == 'y'

    # even if we ignore extra attributes during initialization / validation
    # we never ignore them during assignment
    # instead if extra='ignore' was set (or nothing was set since that's the default)
    # we treat it as if it were extra='forbid'
    with pytest.raises(ValidationError) as exc_info:
        v.validate_assignment(m, 'not_f', 'xyz')
    assert exc_info.value.errors(include_url=False) == [
        {
            'type': 'no_such_attribute',
            'loc': ('not_f',),
            'msg': "Object has no attribute 'not_f'",
            'input': 'xyz',
            'ctx': {'attribute': 'not_f'},
        }
    ]
    assert 'not_f' not in m



from dataclasses import dataclass

from typing_extensions import TypedDict

from pydantic_core import SchemaSerializer, SchemaValidator, core_schema


def test_serialize_as_any_with_models() -> None:
    class Parent:
        x: int

    class Child(Parent):
        y: str

    Parent.__pydantic_core_schema__ = core_schema.model_schema(
        Parent,
        core_schema.model_fields_schema(
            {
                'x': core_schema.model_field(core_schema.int_schema()),
            }
        ),
    )
    Parent.__pydantic_validator__ = SchemaValidator(Parent.__pydantic_core_schema__)
    Parent.__pydantic_serializer__ = SchemaSerializer(Parent.__pydantic_core_schema__)

    Child.__pydantic_core_schema__ = core_schema.model_schema(
        Child,
        core_schema.model_fields_schema(
            {
                'x': core_schema.model_field(core_schema.int_schema()),
                'y': core_schema.model_field(core_schema.str_schema()),
            }
        ),
    )
    Child.__pydantic_validator__ = SchemaValidator(Child.__pydantic_core_schema__)
    Child.__pydantic_serializer__ = SchemaSerializer(Child.__pydantic_core_schema__)

    child = Child.__pydantic_validator__.validate_python({'x': 1, 'y': 'hopefully not a secret'})
    assert Parent.__pydantic_serializer__.to_python(child, serialize_as_any=False) == {'x': 1}
    assert Parent.__pydantic_serializer__.to_python(child, serialize_as_any=True) == {
        'x': 1,
        'y': 'hopefully not a secret',
    }


def test_serialize_as_any_with_dataclass() -> None:
    @dataclass
    class Parent:
        x: int

    class Child(Parent):
        y: str

    Parent.__pydantic_core_schema__ = core_schema.dataclass_schema(
        Parent,
        core_schema.dataclass_args_schema(
            'Parent',
            [
                core_schema.dataclass_field(name='x', schema=core_schema.int_schema()),
            ],
        ),
        ['x'],
    )
    Parent.__pydantic_validator__ = SchemaValidator(Parent.__pydantic_core_schema__)
    Parent.__pydantic_serializer__ = SchemaSerializer(Parent.__pydantic_core_schema__)

    Child.__pydantic_core_schema__ = core_schema.dataclass_schema(
        Child,
        core_schema.dataclass_args_schema(
            'Child',
            [
                core_schema.dataclass_field(name='x', schema=core_schema.int_schema()),
                core_schema.dataclass_field(name='y', schema=core_schema.str_schema()),
            ],
        ),
        ['x', 'y'],
    )
    Child.__pydantic_validator__ = SchemaValidator(Child.__pydantic_core_schema__)
    Child.__pydantic_serializer__ = SchemaSerializer(Child.__pydantic_core_schema__)

    child = Child.__pydantic_validator__.validate_python({'x': 1, 'y': 'hopefully not a secret'})
    assert Parent.__pydantic_serializer__.to_python(child, serialize_as_any=False) == {'x': 1}
    assert Parent.__pydantic_serializer__.to_python(child, serialize_as_any=True) == {
        'x': 1,
        'y': 'hopefully not a secret',
    }


def test_serialize_as_any_with_typeddict() -> None:
    class Parent(TypedDict):
        x: int

    class Child(Parent):
        y: str

    Parent.__pydantic_core_schema__ = core_schema.typed_dict_schema(
        {
            'x': core_schema.typed_dict_field(core_schema.int_schema()),
        }
    )
    Parent.__pydantic_validator__ = SchemaValidator(Parent.__pydantic_core_schema__)
    Parent.__pydantic_serializer__ = SchemaSerializer(Parent.__pydantic_core_schema__)

    Child.__pydantic_core_schema__ = core_schema.typed_dict_schema(
        {
            'x': core_schema.typed_dict_field(core_schema.int_schema()),
            'y': core_schema.typed_dict_field(core_schema.str_schema()),
        }
    )
    Child.__pydantic_validator__ = SchemaValidator(Child.__pydantic_core_schema__)
    Child.__pydantic_serializer__ = SchemaSerializer(Child.__pydantic_core_schema__)

    child = Child.__pydantic_validator__.validate_python({'x': 1, 'y': 'hopefully not a secret'})
    assert Parent.__pydantic_serializer__.to_python(child, serialize_as_any=False) == {'x': 1}
    assert Parent.__pydantic_serializer__.to_python(child, serialize_as_any=True) == {
        'x': 1,
        'y': 'hopefully not a secret',
    }


def test_serialize_as_any_with_unrelated_models() -> None:
    class Parent:
        x: int

    class Other:
        y: str

    Parent.__pydantic_core_schema__ = core_schema.model_schema(
        Parent,
        core_schema.model_fields_schema(
            {
                'x': core_schema.model_field(core_schema.int_schema()),
            }
        ),
    )
    Parent.__pydantic_validator__ = SchemaValidator(Parent.__pydantic_core_schema__)
    Parent.__pydantic_serializer__ = SchemaSerializer(Parent.__pydantic_core_schema__)

    Other.__pydantic_core_schema__ = core_schema.model_schema(
        Other,
        core_schema.model_fields_schema(
            {
                'y': core_schema.model_field(core_schema.str_schema()),
            }
        ),
        config=core_schema.CoreConfig(extra_fields_behavior='allow'),
    )
    Other.__pydantic_validator__ = SchemaValidator(Other.__pydantic_core_schema__)
    Other.__pydantic_serializer__ = SchemaSerializer(Other.__pydantic_core_schema__)

    other = Other.__pydantic_validator__.validate_python({'x': 1, 'y': 'hopefully not a secret'})
    assert Parent.__pydantic_serializer__.to_python(other, serialize_as_any=False) == {}
    # note, without extra='allow', the 'x' field would not be included, as it's not in the schema
    assert Parent.__pydantic_serializer__.to_python(other, serialize_as_any=True) == {
        'x': 1,
        'y': 'hopefully not a secret',
    }



import re
import sys
from enum import Enum, IntEnum, IntFlag

import pytest

from pydantic_core import SchemaError, SchemaValidator, ValidationError, core_schema


def test_plain_enum():
    class MyEnum(Enum):
        a = 1
        b = 2

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())))

    # debug(v)
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(MyEnum.b) is MyEnum.b
    assert v.validate_python(1) is MyEnum.a
    assert v.validate_python(2) is MyEnum.b

    assert v.validate_json('1') is MyEnum.a
    # assert v.validate_json('"1"') is MyEnum.a

    with pytest.raises(ValidationError, match=r'Input should be 1 or 2 \[type=enum, input_value=3, input_type=int\]'):
        v.validate_python(3)

    with pytest.raises(ValidationError, match=r"Input should be 1 or 2 \[type=enum, input_value='1', input_type=str\]"):
        v.validate_python('1')

    assert v.validate_python(MyEnum.a, strict=True) is MyEnum.a

    e = (
        'Input should be an instance of test_plain_enum.<locals>.MyEnum '
        '[type=is_instance_of, input_value=1, input_type=int]'
    )
    with pytest.raises(ValidationError, match=re.escape(e)):
        v.validate_python(1, strict=True)

    assert v.validate_json('1', strict=True) is MyEnum.a
    with pytest.raises(ValidationError, match='type=enum'):
        v.validate_json('"1"', strict=True)

    v_strict = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), strict=True))
    assert v_strict.validate_python(MyEnum.a) is MyEnum.a

    with pytest.raises(ValidationError, match=re.escape(e)):
        v_strict.validate_python(1, strict=True)

    v_strict_f = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), strict=True))
    assert v_strict_f.validate_python(MyEnum.a) is MyEnum.a

    with pytest.raises(ValidationError, match=re.escape(e)):
        v_strict_f.validate_python(1, strict=True)


def test_int_enum():
    class MyEnum(int, Enum):
        a = 1
        b = 2

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), sub_type='int'))

    # debug(v)
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(1) is MyEnum.a
    assert v.validate_python(1.0) is MyEnum.a
    assert v.validate_python('1') is MyEnum.a

    assert v.validate_json('1') is MyEnum.a
    assert v.validate_json('"1"') is MyEnum.a

    with pytest.raises(ValidationError, match=r'Input should be 1 or 2 \[type=enum, input_value=3, input_type=int\]'):
        v.validate_python(3)

    assert v.validate_python(MyEnum.a, strict=True) is MyEnum.a

    e = (
        'Input should be an instance of test_int_enum.<locals>.MyEnum '
        '[type=is_instance_of, input_value=1, input_type=int]'
    )
    with pytest.raises(ValidationError, match=re.escape(e)):
        v.validate_python(1, strict=True)

    assert v.validate_json('1', strict=True) is MyEnum.a

    with pytest.raises(ValidationError, match=r"Input should be 1 or 2 \[type=enum, input_value='1', input_type=str\]"):
        v.validate_json('"1"', strict=True)


def test_str_enum():
    class MyEnum(str, Enum):
        a = 'x'
        b = 'y'

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), sub_type='str'))

    # debug(v)
    assert v.validate_python('x') is MyEnum.a
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(b'x') is MyEnum.a

    assert v.validate_json('"x"') is MyEnum.a

    with pytest.raises(
        ValidationError, match=r"Input should be 'x' or 'y' \[type=enum, input_value='a', input_type=str\]"
    ):
        v.validate_python('a')

    assert v.validate_python(MyEnum.a, strict=True) is MyEnum.a

    e = (
        'Input should be an instance of test_str_enum.<locals>.MyEnum '
        "[type=is_instance_of, input_value='x', input_type=str]"
    )
    with pytest.raises(ValidationError, match=re.escape(e)):
        v.validate_python('x', strict=True)
    assert v.validate_json('"x"', strict=True) is MyEnum.a


def test_float_enum():
    class MyEnum(float, Enum):
        a = 1.5
        b = 2.5
        c = 3.0

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), sub_type='float'))

    # debug(v)
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(1.5) is MyEnum.a
    assert v.validate_python('1.5') is MyEnum.a
    assert v.validate_python(3) is MyEnum.c

    assert v.validate_json('1.5') is MyEnum.a
    assert v.validate_json('"1.5"') is MyEnum.a

    e = r'Input should be 1.5, 2.5 or 3.0 \[type=enum, input_value=4.0, input_type=float\]'
    with pytest.raises(ValidationError, match=e):
        v.validate_python(4.0)

    assert v.validate_python(MyEnum.a, strict=True) is MyEnum.a

    e = (
        'Input should be an instance of test_float_enum.<locals>.MyEnum '
        '[type=is_instance_of, input_value=1.5, input_type=float]'
    )
    with pytest.raises(ValidationError, match=re.escape(e)):
        v.validate_python(1.5, strict=True)

    assert v.validate_json('1.5', strict=True) is MyEnum.a

    with pytest.raises(ValidationError, match='type=enum'):
        v.validate_json('"3.0"', strict=True)


def test_enum_missing():
    class MyEnum(Enum):
        a = 1
        b = 2

        @classmethod
        def _missing_(cls, v):
            return cls.b

    assert MyEnum(1) is MyEnum.a
    assert MyEnum(2) is MyEnum.b
    assert MyEnum(3) is MyEnum.b

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), missing=MyEnum._missing_))

    # debug(v)
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(1) is MyEnum.a
    assert v.validate_python(2) is MyEnum.b
    assert v.validate_python(3) is MyEnum.b

    assert v.validate_json('1') is MyEnum.a
    assert v.validate_json('3') is MyEnum.b


def test_enum_missing_none():
    class MyEnum(Enum):
        a = 1
        b = 2

        @classmethod
        def _missing_(cls, v):
            return None

    assert MyEnum(1) is MyEnum.a
    assert MyEnum(2) is MyEnum.b
    with pytest.raises(ValueError, match='3 is not a valid'):
        MyEnum(3)

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), missing=MyEnum._missing_))

    # debug(v)
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(1) is MyEnum.a
    with pytest.raises(ValidationError, match=r'Input should be 1 or 2 \[type=enum, input_value=3, input_type=int\]'):
        v.validate_python(3)

    assert v.validate_json('1') is MyEnum.a
    with pytest.raises(ValidationError, match=r'Input should be 1 or 2 \[type=enum, input_value=3, input_type=int\]'):
        v.validate_json('3')


def test_enum_missing_wrong():
    class MyEnum(Enum):
        a = 1
        b = 2

        @classmethod
        def _missing_(cls, v):
            return 'foobar'

    assert MyEnum(1) is MyEnum.a
    assert MyEnum(2) is MyEnum.b
    # different error from pypy
    if sys.implementation.name == 'pypy':
        e = "returned 'foobar' instead of None or a valid member"
    else:
        e = "error in MyEnum._missing_: returned 'foobar' instead of None or a valid member"
    with pytest.raises(TypeError, match=e):
        MyEnum(3)

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), missing=MyEnum._missing_))
    with pytest.raises(TypeError, match=e):
        v.validate_python(3)


def test_enum_exactness():
    class MyEnum(int, Enum):
        a = 1
        b = 2

    v = SchemaValidator(
        core_schema.union_schema(
            [
                core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values()), missing=MyEnum._missing_),
                core_schema.int_schema(),
            ],
        )
    )
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(1) == 1
    assert v.validate_python(1) is not MyEnum.a


def test_plain_enum_lists():
    class MyEnum(Enum):
        a = [1]
        b = [2]

    assert MyEnum([1]) is MyEnum.a
    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())))
    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python([1]) is MyEnum.a
    assert v.validate_python([2]) is MyEnum.b


def test_plain_enum_empty():
    class MyEnum(Enum):
        pass

    with pytest.raises(SchemaError, match='`members` should have length > 0'):
        SchemaValidator(core_schema.enum_schema(MyEnum, []))


def test_enum_with_str_subclass() -> None:
    class MyEnum(Enum):
        a = 'a'
        b = 'b'

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())))

    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python('a') is MyEnum.a

    class MyStr(str):
        pass

    assert v.validate_python(MyStr('a')) is MyEnum.a
    with pytest.raises(ValidationError):
        v.validate_python(MyStr('a'), strict=True)


def test_enum_with_int_subclass() -> None:
    class MyEnum(Enum):
        a = 1
        b = 2

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())))

    assert v.validate_python(MyEnum.a) is MyEnum.a
    assert v.validate_python(1) is MyEnum.a

    class MyInt(int):
        pass

    assert v.validate_python(MyInt(1)) is MyEnum.a
    with pytest.raises(ValidationError):
        v.validate_python(MyInt(1), strict=True)


def test_validate_float_for_int_enum() -> None:
    class MyEnum(int, Enum):
        a = 1
        b = 2

    v = SchemaValidator(core_schema.enum_schema(MyEnum, list(MyEnum.__members__.values())))

    assert v.validate_python(1.0) is MyEnum.a


def test_missing_error_converted_to_val_error() -> None:
    class MyFlags(IntFlag):
        OFF = 0
        ON = 1

    v = SchemaValidator(
        core_schema.with_default_schema(
            schema=core_schema.enum_schema(MyFlags, list(MyFlags.__members__.values())), default=MyFlags.OFF
        )
    )

    assert v.validate_python(MyFlags.OFF) is MyFlags.OFF
    assert v.validate_python(0) is MyFlags.OFF

    with pytest.raises(ValidationError):
        v.validate_python(None)


def test_big_int():
    class ColorEnum(IntEnum):
        GREEN = 1 << 63
        BLUE = 1 << 64

    v = SchemaValidator(
        core_schema.with_default_schema(schema=core_schema.enum_schema(ColorEnum, list(ColorEnum.__members__.values())))
    )

    assert v.validate_python(ColorEnum.GREEN) is ColorEnum.GREEN
    assert v.validate_python(1 << 63) is ColorEnum.GREEN



import pytest

from pydantic_core import SchemaSerializer, SchemaValidator, core_schema

from ..conftest import plain_repr


def test_chain():
    s = SchemaSerializer(core_schema.chain_schema([core_schema.str_schema(), core_schema.int_schema()]))

    # insert_assert(plain_repr(s))
    assert plain_repr(s) == 'SchemaSerializer(serializer=Int(IntSerializer),definitions=[])'

    assert s.to_python(1) == 1
    assert s.to_json(1) == b'1'


def test_function_plain():
    s = SchemaSerializer(core_schema.with_info_plain_validator_function(lambda v, info: v + 1))
    # can't infer the type from plain function validators
    # insert_assert(plain_repr(s))
    assert plain_repr(s) == 'SchemaSerializer(serializer=Any(AnySerializer),definitions=[])'


def test_function_before():
    s = SchemaSerializer(
        core_schema.with_info_before_validator_function(lambda v, info: v + 1, core_schema.int_schema())
    )
    # insert_assert(plain_repr(s))
    assert plain_repr(s) == 'SchemaSerializer(serializer=Int(IntSerializer),definitions=[])'


def test_function_after():
    s = SchemaSerializer(
        core_schema.with_info_after_validator_function(lambda v, info: v + 1, core_schema.int_schema())
    )
    # insert_assert(plain_repr(s))
    assert plain_repr(s) == 'SchemaSerializer(serializer=Int(IntSerializer),definitions=[])'


def test_lax_or_strict():
    s = SchemaSerializer(core_schema.lax_or_strict_schema(core_schema.int_schema(), core_schema.str_schema()))
    # insert_assert(plain_repr(s))
    assert plain_repr(s) == 'SchemaSerializer(serializer=Str(StrSerializer),definitions=[])'

    assert s.to_json('abc') == b'"abc"'
    with pytest.warns(
        UserWarning, match='Expected `str` but got `int` with value `123` - serialized value may not be as expected'
    ):
        assert s.to_json(123) == b'123'


def test_lax_or_strict_custom_ser():
    s = SchemaSerializer(
        core_schema.lax_or_strict_schema(
            core_schema.int_schema(),
            core_schema.str_schema(),
            serialization=core_schema.format_ser_schema('^5s', when_used='always'),
        )
    )

    assert s.to_python('abc') == ' abc '
    assert s.to_python('abc', mode='json') == ' abc '
    assert s.to_json('abc') == b'" abc "'


def test_serialize_with_extra_on_superclass() -> None:
    class Parent:
        x: int

    class Other(Parent):
        y: str

    Parent.__pydantic_core_schema__ = core_schema.model_schema(
        Parent,
        core_schema.model_fields_schema(
            {
                'x': core_schema.model_field(core_schema.int_schema()),
            }
        ),
        config=core_schema.CoreConfig(extra_fields_behavior='allow'),
    )
    Parent.__pydantic_validator__ = SchemaValidator(Parent.__pydantic_core_schema__)
    Parent.__pydantic_serializer__ = SchemaSerializer(Parent.__pydantic_core_schema__)

    Other.__pydantic_core_schema__ = core_schema.model_schema(
        Other,
        core_schema.model_fields_schema(
            {
                'x': core_schema.model_field(core_schema.int_schema()),
                'y': core_schema.model_field(core_schema.str_schema()),
            }
        ),
        config=core_schema.CoreConfig(extra_fields_behavior='forbid'),
    )
    Other.__pydantic_validator__ = SchemaValidator(Other.__pydantic_core_schema__)
    Other.__pydantic_serializer__ = SchemaSerializer(Other.__pydantic_core_schema__)

    other = Other.__pydantic_validator__.validate_python({'x': 1, 'y': 'some string'})
    assert Parent.__pydantic_serializer__.to_python(other) == {'x': 1}



import json
from enum import IntEnum

import pytest

from pydantic_core import SchemaSerializer, core_schema

try:
    import numpy
except ImportError:
    numpy = None


class IntSubClass(int):
    pass


class MyIntEnum(IntEnum):
    one = 1
    two = 2


class FloatSubClass(float):
    pass


# A number well outside of i64 range
_BIG_NUMBER_BYTES = b'1' + (b'0' * 40)


@pytest.mark.parametrize('custom_type_schema', [None, 'any'])
@pytest.mark.parametrize(
    'schema_type,value,expected_python,expected_json',
    [
        ('int', 1, 1, b'1'),
        ('int', int(_BIG_NUMBER_BYTES), int(_BIG_NUMBER_BYTES), _BIG_NUMBER_BYTES),
        ('bool', True, True, b'true'),
        ('bool', False, False, b'false'),
        ('float', 1.0, 1.0, b'1.0'),
        ('float', 42.31415, 42.31415, b'42.31415'),
        ('none', None, None, b'null'),
        ('int', IntSubClass(42), IntSubClass(42), b'42'),
        ('int', MyIntEnum.one, MyIntEnum.one, b'1'),
        ('float', FloatSubClass(42), FloatSubClass(42), b'42.0'),
    ],
)
def test_simple_serializers(schema_type, value, expected_python, expected_json, custom_type_schema):
    if custom_type_schema is None:
        schema = {'type': schema_type}
    else:
        schema = {'type': custom_type_schema}

    s = SchemaSerializer(schema)
    v = s.to_python(value)
    assert v == expected_python
    assert type(v) == type(expected_python)

    assert s.to_json(value) == expected_json

    v_json = s.to_python(value, mode='json')
    v_json_expected = json.loads(expected_json)
    assert v_json == v_json_expected
    assert type(v_json) == type(v_json_expected)


def test_int_to_float():
    """
    See https://github.com/pydantic/pydantic-core/pull/866
    """
    s = SchemaSerializer(core_schema.float_schema())
    v_plain = s.to_python(1)
    assert v_plain == 1
    assert type(v_plain) == int

    v_plain_subclass = s.to_python(IntSubClass(1))
    assert v_plain_subclass == IntSubClass(1)
    assert type(v_plain_subclass) == IntSubClass

    v_json = s.to_python(1, mode='json')
    assert v_json == 1.0
    assert type(v_json) == float

    v_json_subclass = s.to_python(IntSubClass(1), mode='json')
    assert v_json_subclass == 1
    assert type(v_json_subclass) == float

    assert s.to_json(1) == b'1.0'
    assert s.to_json(IntSubClass(1)) == b'1.0'


def test_int_to_float_key():
    """
    See https://github.com/pydantic/pydantic-core/pull/866
    """
    s = SchemaSerializer(core_schema.dict_schema(core_schema.float_schema(), core_schema.float_schema()))
    v_plain = s.to_python({1: 1})
    assert v_plain == {1: 1}
    assert type(list(v_plain.keys())[0]) == int
    assert type(v_plain[1]) == int

    v_json = s.to_python({1: 1}, mode='json')
    assert v_json == {'1': 1.0}
    assert type(v_json['1']) == float

    assert s.to_json({1: 1}) == b'{"1":1.0}'


@pytest.mark.parametrize('schema_type', ['int', 'bool', 'float', 'none'])
def test_simple_serializers_fallback(schema_type):
    s = SchemaSerializer({'type': schema_type})
    with pytest.warns(
        UserWarning,
        match=f'Expected `{schema_type}` but got `list` with value `\\[1, 2, 3\\]` - serialized value may not be as expected',
    ):
        assert s.to_python([1, 2, 3]) == [1, 2, 3]

    with pytest.warns(
        UserWarning,
        match=f"Expected `{schema_type}` but got `list` with value `\\[1, 2, b'bytes'\\]` - serialized value may not be as expected",
    ):
        assert s.to_python([1, 2, b'bytes'], mode='json') == [1, 2, 'bytes']

    with pytest.warns(
        UserWarning,
        match=f'Expected `{schema_type}` but got `list` with value `\\[1, 2, 3\\]` - serialized value may not be as expected',
    ):
        assert s.to_json([1, 2, 3]) == b'[1,2,3]'


@pytest.mark.skipif(numpy is None, reason='numpy is not installed')
def test_numpy():
    s = SchemaSerializer(core_schema.float_schema())
    v = s.to_python(numpy.float64(1.0))
    assert v == 1.0
    assert type(v) == numpy.float64

    v = s.to_python(numpy.float64(1.0), mode='json')
    assert v == 1.0
    assert type(v) == float

    assert s.to_json(numpy.float64(1.0)) == b'1.0'


@pytest.mark.parametrize(
    'value,expected_json,config',
    [
        # default values of ser_json_inf_nan
        (float('inf'), 'null', {}),
        (float('-inf'), 'null', {}),
        (float('nan'), 'null', {}),
        # explicit values of ser_json_inf_nan
        (float('inf'), 'null', {'ser_json_inf_nan': 'null'}),
        (float('-inf'), 'null', {'ser_json_inf_nan': 'null'}),
        (float('nan'), 'null', {'ser_json_inf_nan': 'null'}),
        (float('inf'), 'Infinity', {'ser_json_inf_nan': 'constants'}),
        (float('-inf'), '-Infinity', {'ser_json_inf_nan': 'constants'}),
        (float('nan'), 'NaN', {'ser_json_inf_nan': 'constants'}),
        (float('inf'), '"Infinity"', {'ser_json_inf_nan': 'strings'}),
        (float('-inf'), '"-Infinity"', {'ser_json_inf_nan': 'strings'}),
        (float('nan'), '"NaN"', {'ser_json_inf_nan': 'strings'}),
    ],
)
def test_float_inf_and_nan_serializers(value, expected_json, config):
    s = SchemaSerializer(core_schema.float_schema(), config)

    # Python can represent these values without needing any changes
    assert s.to_python(value) is value
    assert s.to_python(value, mode='json') is value

    # Serialized JSON value respects the ser_json_inf_nan setting
    assert s.to_json(value).decode() == expected_json
